from .Table import Table
from .Version import Version
import os

from .Query import Query
from .Base import Base
from urllib.parse import quote as quote_uri
import re

from ..common import exceptions
from ..common.api_request import make_request, make_paginated_request


class Dataset(Base):
    def __init__(
        self,
        name,
        *,
        version=None,
        user=None,
        organization=None,
        properties=None,
    ):
        if not user and not organization:
            from .User import User
            from .Organization import Organization

            if len(name.split(".")) == 2:
                username = name.split(".")[0]
                name = name.split(".")[1]
                organization = Organization(username)
                user = User(username)
            elif os.getenv("REDIVIS_DEFAULT_USER"):
                user = User(os.getenv("REDIVIS_DEFAULT_USER"))
            elif os.getenv("REDIVIS_DEFAULT_ORGANIZATION"):
                organization = Organization(os.getenv("REDIVIS_DEFAULT_ORGANIZATION"))
            else:
                raise exceptions.ValueError(
                    "Invalid dataset specifier, must be the fully qualified reference if no owner is specified"
                )

        self.name = name.split(":")[0]
        self.user = user
        self.organization = organization
        reference_id = ""
        reference_id_split = re.sub(
            r":(v\d+[._]\d+|current|next|sample)", "", name
        ).split(":")
        if len(reference_id_split) > 1:
            reference_id = reference_id_split[1]

        if reference_id:
            reference_id = f":{reference_id}"

        if not version:
            version_search = re.search(r":(v\d+[._]\d+|current|next)", name)
            if version_search:
                version = version_search[1]

        if (
            version
            and version != "current"
            and version != "next"
            and not version.lower().startswith("v")
        ):
            version = f"v{version}"

        self.version_tag = version
        self.scoped_reference = (properties or {}).get(
            "scopedReference",
            f"{self.name}{reference_id}{f':{version}' if version else ''}",
        )
        self.qualified_reference = (properties or {}).get(
            "qualifiedReference",
            f"{(self.organization or self.user).name}.{self.scoped_reference}",
        )

        self.uri = f"/datasets/{quote_uri(self.qualified_reference, '')}"
        self.properties = properties

    def __repr__(self):
        return f"<Dataset {self.qualified_reference}>"

    def _rectify_ambiguous_owner(self):
        if self.user and self.organization:
            if (self.properties or {}).get("owner"):
                if self.properties.get("owner")["kind"] == "user":
                    self.organization = None
                else:
                    self.user = None
            elif self.organization.exists():
                self.user = None
            else:
                self.organization = None

    def create(self, *, public_access_level="none", description=None):
        self._rectify_ambiguous_owner()
        if self.organization:
            path = f"/organizations/{self.organization.name}/datasets"
        else:
            path = f"/users/{self.user.name}/datasets"

        properties = make_request(
            method="POST",
            path=path,
            payload={
                "name": self.name,
                "publicAccessLevel": public_access_level,
                "description": description,
            },
        )
        update_properties(self, properties)
        return self

    def create_next_version(self, *, if_not_exists=False):
        if not self.properties or "nextVersion" not in self.properties:
            self.get()

        if not self.properties["nextVersion"]:
            make_request(method="POST", path=f"{self.uri}/versions")
        elif not if_not_exists:
            raise exceptions.ValueError(
                f"Next version already exists at {self.properties['nextVersion']['datasetUri']}. To avoid this error, set argument if_not_exists to True"
            )

        self._rectify_ambiguous_owner()
        next_version_dataset = Dataset(
            name=self.scoped_reference,
            user=self.user,
            organization=self.organization,
            version="next",
        )

        return next_version_dataset.get()

    def delete(self):
        make_request(
            method="DELETE",
            path=self.uri,
        )
        return

    def exists(self):
        try:
            make_request(method="HEAD", path=self.uri)
            return True
        except exceptions.NotFoundError:
            return False

    def get(self):
        properties = make_request(method="GET", path=self.uri)
        update_properties(self, properties)
        return self

    def list_tables(self, max_results=None):
        tables = make_paginated_request(
            path=f"{self.uri}/tables", page_size=100, max_results=max_results
        )
        return [
            Table(table["name"], dataset=self, properties=table) for table in tables
        ]

    def list_versions(self, max_results=None):
        versions = make_paginated_request(
            path=f"{self.uri}/versions", page_size=100, max_results=max_results
        )
        return [
            Version(version["tag"], dataset=self, properties=version)
            for version in versions
        ]

    def query(self, query):
        return Query(query, default_dataset=self.qualified_reference)

    def release(self, *, release_notes=None):
        version_res = make_request(
            method="POST",
            path=f"{self.uri}/versions/next/release",
            payload={"releaseNotes": release_notes},
        )
        self.uri = version_res["datasetUri"]
        self.get()
        return self

    def unrelease(self):
        version_res = make_request(
            method="POST",
            path=f"{self.uri}/versions/current/unrelease",
        )
        self.uri = version_res["datasetUri"]
        self.get()
        return self

    def version(self, tag=None):
        if not self.version_tag and not tag:
            self.get()

        tag = self.version_tag if tag is None else tag
        return Version(tag, dataset=self)

    def previous_version(self):
        if not self.version_tag:
            self.get()
        return Version(self.version_tag, dataset=self).previous_version().dataset()

    def next_version(self):
        if not self.version_tag:
            self.get()
        return Version(self.version_tag, dataset=self).next_version().dataset()

    def table(self, name):
        return Table(name, dataset=self)

    def add_labels(self, labels):
        self.get()
        self.update(
            labels=list(
                set(self.properties.get("labels", []))
                | set(label.lower() for label in labels)
            )
        )
        return self

    def remove_labels(self, labels):
        self.get()
        self.update(
            labels=list(
                set(self.properties.get("labels", []))
                - set(label.lower() for label in labels)
            )
        )
        return self

    def update(
        self, *, name=None, public_access_level=None, description=None, labels=None
    ):
        payload = {}
        if name:
            payload["name"] = name
        if public_access_level:
            payload["publicAccessLevel"] = public_access_level
        if description is not None:
            payload["description"] = description
        if labels is not None:
            payload["labels"] = labels

        res = make_request(
            method="PATCH",
            path=self.uri,
            payload=payload,
        )
        update_properties(self, res)
        return self

    def update_variables(self, variables):
        make_request(
            method="PATCH",
            path=f"{self.uri}/variables",
            payload={"variables": variables},
        )


def update_properties(instance, properties):
    instance.properties = properties
    instance.qualified_reference = properties["qualifiedReference"]
    instance.scoped_reference = properties["scopedReference"]
    instance.name = properties["name"]
    instance.uri = properties["uri"]
    instance._rectify_ambiguous_owner()



import concurrent.futures
import uuid
import os
import pathlib
import time
from requests import RequestException
from urllib3.exceptions import HTTPError
from contextlib import closing, nullcontext

from ..common import exceptions
from tqdm.auto import tqdm
import shutil
from .util import get_tempdir
from .api_request import make_request
from threading import Event

MAX_PARALLELIZATION = 8


class RedivisArrowIterator:
    def __init__(self, streams, mapped_variables, progressbar, coerce_schema):
        self.streams = streams
        self.mapped_variables = mapped_variables
        self.progressbar = progressbar
        self.coerce_schema = coerce_schema
        self.current_stream_index = 0
        self.current_offset = 0
        self.retry_count = 0
        self.__get_next_reader__()

    def __get_next_reader__(self, offset=0):
        import pyarrow

        try:
            self.current_offset = offset
            # TODO: this won't get closed properly if the iterator is not fully consumed
            arrow_response = make_request(
                method="get",
                path=f'/readStreams/{self.streams[self.current_stream_index]["id"]}?offset={offset}',
                stream=True,
                parse_response=False,
            )
            self.current_record_batch_reader = pyarrow.ipc.RecordBatchStreamReader(
                arrow_response.raw
            )
            if self.coerce_schema:
                self.variables_in_stream = list(
                    map(
                        lambda field_name: next(
                            x for x in self.mapped_variables if x["name"] == field_name
                        ),
                        self.current_record_batch_reader.schema.names,
                    )
                )
                self.output_schema = pyarrow.schema(
                    map(variable_to_field, self.variables_in_stream)
                )
            else:
                self.output_schema = self.current_record_batch_reader.schema
        except (RequestException, HTTPError) as e:
            self.retry_count = self.retry_count + 1
            if self.retry_count > 10:
                raise exceptions.NetworkError(
                    message=f"Download connection failed after {self.retry_count} retries.",
                    original_exception=e,
                ) from e

            time.sleep(self.retry_count)
            return self.__get_next_reader__(self.current_offset)

    def __iter__(self):
        return self

    def __next__(self):
        import pyarrow

        try:
            batch = self.current_record_batch_reader.read_next_batch()
            if self.coerce_schema:
                batch = pyarrow.RecordBatch.from_arrays(
                    list(
                        map(
                            coerce_arrow_array,
                            batch.columns,
                            self.variables_in_stream,
                        )
                    ),
                    schema=self.output_schema,
                )

            if self.progressbar is not None:
                self.progressbar.update(batch.num_rows)

            self.current_offset += batch.num_rows
            self.retry_count = 0
            return batch
        except StopIteration:
            if self.current_stream_index == len(self.streams) - 1:
                if self.progressbar:
                    self.progressbar.close()
                raise StopIteration
            else:
                self.current_stream_index += 1
                self.__get_next_reader__()
                return self.__next__()
        except (RequestException, HTTPError) as e:
            self.retry_count = self.retry_count + 1
            if self.retry_count > 10:
                raise exceptions.NetworkError(
                    message=f"A network error occurred. Download connection failed after {self.retry_count} retries.",
                    original_exception=e,
                ) from e
            time.sleep(self.retry_count)
            self.__get_next_reader__(self.current_offset)
            return self.__next__()


def list_rows(
    *,
    uri,
    output_type="dataframe",
    max_results=None,
    selected_variables=None,
    mapped_variables=None,
    progress=True,
    coerce_schema=False,
    batch_preprocessor=None,
    instance=None,
    use_export_api=False,
    max_parallelization=os.cpu_count(),
):
    import pyarrow
    import pyarrow.dataset as pyarrow_dataset  # need to import separately, it's not on the pyarrow import
    from ..classes.Table import Table
    from ..classes.ReadStream import ReadStream

    progressbar = None

    if isinstance(instance, ReadStream):
        read_session = {
            "streams": [{"id": instance.id}],
            "numRows": instance.properties.get("estimatedRows", 0),
        }
        max_parallelization = 1
    else:
        if max_parallelization < 1:
            raise exceptions.ValueError("max_parallelization must be greater than 0")

        pyarrow.set_cpu_count(max_parallelization)
        pyarrow.set_io_thread_count(max_parallelization)

        use_export_api = (
            use_export_api
            and isinstance(instance, Table)
            and output_type != "arrow_iterator"
            and selected_variables is None
            and batch_preprocessor is None
            and max_results is None
        )
        payload = {
            "requestedStreamCount": min(MAX_PARALLELIZATION, max_parallelization)
        }

        if max_results is not None:
            payload["maxResults"] = max_results

        if selected_variables is not None:
            payload["selectedVariables"] = selected_variables

        if not use_export_api:
            read_session = make_request(
                method="post",
                path=f"{uri}/readSessions",
                parse_response=True,
                payload=payload,
            )

    if progress:
        progressbar = tqdm(total=read_session["numRows"], leave=False, mininterval=0.1)

    if output_type == "arrow_iterator":
        return RedivisArrowIterator(
            streams=read_session["streams"],
            mapped_variables=mapped_variables,
            progressbar=progressbar,
            coerce_schema=coerce_schema,
        )

    folder = None
    folder_path = None
    # get the absolute folder path, as a string
    # We need to always write to disk if we're doing things in parallel,
    # because we can't efficiently copy results between processes when working in parallel
    if (
        use_export_api
        or output_type in ["arrow_dataset", "dask_dataframe", "polars_lazyframe"]
        or (len(read_session["streams"]) > 1 and max_parallelization > 1)
    ):
        folder = pathlib.Path().joinpath(
            get_tempdir(),
            "tables",
            f"{uuid.uuid4()}",
        )
        folder_path = str(folder.absolute())

    try:
        arrow_dataset = None
        all_batches = []
        if use_export_api:
            instance.download(folder_path + "/", format="parquet", progress=progress)
        else:
            if folder_path is not None:
                # create the folder, if it doesn't exist
                folder.mkdir(parents=True, exist_ok=True)

            # Use download_state to notify worker threads when to quit.
            # See: https://stackoverflow.com/a/29237343/101923
            cancel_event = Event()

            # Code to use multiprocess... would simplify exiting on stop, but progress doesn't currently work
            # with concurrent.futures.ProcessPoolExecutor(max_workers=len(read_session["streams"]), mp_context=mp.get_context('fork')) as executor:

            # See https://github.com/googleapis/python-bigquery/blob/main/google/cloud/bigquery/_pandas_helpers.py#L920
            futures = []
            if len(read_session["streams"]):
                with concurrent.futures.ThreadPoolExecutor(
                    max_workers=min(max_parallelization, len(read_session["streams"]))
                ) as executor:
                    futures = [
                        executor.submit(
                            process_stream,
                            stream,
                            folder_path,
                            mapped_variables,
                            coerce_schema,
                            progressbar,
                            batch_preprocessor,
                            cancel_event,
                        )
                        for stream in read_session["streams"]
                    ]

                    not_done = futures

                    try:
                        while not_done and not cancel_event.is_set():
                            # next line 'sleeps' this main thread, letting the thread pool run
                            freshly_done, not_done = concurrent.futures.wait(
                                not_done, timeout=0.2
                            )
                            for future in freshly_done:
                                # Call result() on any finished threads to raise any exceptions encountered.
                                res = future.result()
                                if folder_path is None and res:
                                    all_batches.extend(res)
                    finally:
                        cancel_event.set()
                        # Shutdown all background threads, now that they should know to exit early.
                        executor.shutdown(wait=True, cancel_futures=True)

            if progressbar:
                progressbar.close()

        if folder_path is None:
            if all_batches:
                return pyarrow.Table.from_batches(all_batches)

            # No batches were returned; construct an empty table with the expected schema
            schema = pyarrow.schema(map(variable_to_field, mapped_variables))
            return pyarrow.Table.from_batches([], schema=schema)
        elif use_export_api:
            if output_type == "polars_lazyframe":
                import polars

                return polars.scan_parquet(f"{folder_path}/*")
            elif output_type == "dask_dataframe":
                import dask.dataframe as dd

                return dd.read_parquet(folder_path, dtype_backend="pyarrow")

            arrow_dataset = pyarrow_dataset.dataset(folder_path, format="parquet")
            if output_type == "arrow_dataset":
                return arrow_dataset
            else:
                arrow_table = arrow_dataset.to_table()
                shutil.rmtree(folder_path, ignore_errors=True)
                return arrow_table
        else:
            if output_type == "polars_lazyframe":
                import polars

                return polars.scan_ipc(f"{folder_path}/*", memory_map=True)

            arrow_dataset = pyarrow_dataset.dataset(
                folder_path,
                format="feather",
                schema=(
                    pyarrow.schema(map(variable_to_field, mapped_variables))
                    if batch_preprocessor is None
                    else None
                ),
            )

        if output_type == "arrow_dataset":
            return arrow_dataset
        elif output_type == "dask_dataframe":
            import dask.dataframe as dd

            # TODO: simplify once dask supports reading from feather: https://github.com/dask/dask/issues/6865
            # Make sure we no longer remove the folder in the finally clause after making this change
            # Create the Parquet base directory
            parquet_base_dir = str(
                pathlib.Path()
                .joinpath(
                    get_tempdir(),
                    "tables",
                    f"{uuid.uuid4()}",
                )
                .absolute()
            )
            pyarrow_dataset.write_dataset(
                arrow_dataset, parquet_base_dir, format="parquet"
            )
            shutil.rmtree(folder_path, ignore_errors=True)
            return dd.read_parquet(parquet_base_dir, dtype_backend="pyarrow")
        else:
            arrow_table = arrow_dataset.to_table()
            shutil.rmtree(folder_path, ignore_errors=True)
            return arrow_table
    finally:
        if (
            folder_path
            and output_type != "arrow_dataset"
            and output_type != "polars_lazyframe"
        ):
            shutil.rmtree(folder_path, ignore_errors=True)


def variable_to_field(variable):
    import pyarrow

    if variable["type"] == "string" or variable["type"] == "geography":
        return pyarrow.field(variable["name"], pyarrow.string())
    elif variable["type"] == "integer":
        return pyarrow.field(variable["name"], pyarrow.int64())
    elif variable["type"] == "float":
        return pyarrow.field(variable["name"], pyarrow.float64())
    elif variable["type"] == "date":
        return pyarrow.field(variable["name"], pyarrow.date32())
    elif variable["type"] == "dateTime":
        return pyarrow.field(variable["name"], pyarrow.timestamp("us"))
    elif variable["type"] == "time":
        return pyarrow.field(variable["name"], pyarrow.time64("us"))
    elif variable["type"] == "boolean":
        return pyarrow.field(variable["name"], pyarrow.bool_())


# If streaming from a dataset, data types _may_ be incorrect. We need to check and convert if possible.
def coerce_arrow_array(pyarrow_array, variable):
    import pyarrow

    if variable["type"] == "string" or variable["type"] == "geography":
        return pyarrow_array
    elif variable["type"] == "integer":
        if pyarrow_array.type == "int64":
            return pyarrow_array
        else:
            return pyarrow.compute.cast(pyarrow_array, pyarrow.int64())
    elif variable["type"] == "float":
        if pyarrow_array.type == "double":
            return pyarrow_array
        else:
            return pyarrow.compute.cast(pyarrow_array, pyarrow.float64())
    elif variable["type"] == "date":
        if pyarrow_array.type == "date32[day]":
            return pyarrow_array
        else:
            return pyarrow.compute.cast(pyarrow_array, pyarrow.date32())
    elif variable["type"] == "dateTime":
        if pyarrow_array.type == "timestamp[us]":
            return pyarrow_array
        else:
            return pyarrow.compute.cast(pyarrow_array, pyarrow.timestamp("us"))
    elif variable["type"] == "time":
        if pyarrow_array.type == "time64[us]":
            return pyarrow_array
        else:
            # Hopefully someday this is supported. Until then, need to do the workaround below
            # return pyarrow.compute.cast(pyarrow_array, pyarrow.time64("us"))
            return pyarrow.compute.cast(
                pyarrow.compute.cast(
                    pyarrow.compute.utf8_replace_slice(
                        pyarrow_array, start=0, stop=0, replacement="2020-01-01T"
                    ),
                    pyarrow.timestamp("us"),
                ),
                pyarrow.time64("us"),
            )
    elif variable["type"] == "boolean":
        if pyarrow_array.type == "bool":
            return pyarrow_array
        else:
            return pyarrow.compute.cast(pyarrow_array, pyarrow.bool_())


def process_stream(
    stream,
    folder_path,
    mapped_variables,
    coerce_schema,
    progressbar,
    batch_preprocessor,
    cancel_event,
    offset=0,
    retry_count=0,
):
    try:
        import pyarrow

        with closing(
            make_request(
                method="get",
                path=f'/readStreams/{stream["id"]}?offset={offset}',
                stream=True,
                parse_response=False,
            )
        ) as arrow_response:
            record_batches = [] if folder_path is None else None
            has_content = False
            # create the os_file path
            os_file = (
                pathlib.Path(folder_path).joinpath(f"{stream['id']}.feather").absolute()
                if folder_path is not None
                else None
            )
            file_context = (
                pyarrow.OSFile(str(os_file), mode="wb")
                if folder_path is not None
                else nullcontext()
            )
            with file_context as f, pyarrow.ipc.RecordBatchStreamReader(
                arrow_response.raw
            ) as reader:
                if coerce_schema:
                    variables_in_stream = list(
                        map(
                            lambda field_name: next(
                                x
                                for x in mapped_variables
                                if x["name"].lower() == field_name.lower()
                            ),
                            reader.schema.names,
                        )
                    )

                    output_schema = pyarrow.schema(
                        map(variable_to_field, variables_in_stream)
                    )
                else:
                    output_schema = reader.schema
                writer = None
                for batch in reader:
                    # exit out of thread
                    if cancel_event.is_set():
                        has_content = False
                        break

                    if coerce_schema:
                        batch = pyarrow.RecordBatch.from_arrays(
                            list(
                                map(
                                    coerce_arrow_array,
                                    batch.columns,
                                    variables_in_stream,
                                )
                            ),
                            schema=output_schema,
                        )

                    num_rows = batch.num_rows
                    offset += num_rows
                    if batch_preprocessor:
                        batch = batch_preprocessor(batch)

                    if batch is not None:
                        has_content = True
                        if folder_path is None:
                            record_batches.append(batch)
                        else:
                            if writer is None:
                                writer = pyarrow.ipc.RecordBatchFileWriter(
                                    f,
                                    (
                                        output_schema
                                        if batch_preprocessor is None
                                        else batch.schema
                                    ),
                                )

                            writer.write_batch(batch)

                    if progressbar is not None:
                        progressbar.update(num_rows)

                if writer is not None:
                    writer.close()

            if folder_path is None:
                return record_batches
            elif not has_content:
                os.remove(os_file)
    except (RequestException, HTTPError) as e:
        if retry_count >= 10:
            raise exceptions.NetworkError(
                message=f"A network error occurred. Stream rows connection failed after {retry_count} retries.",
                original_exception=e,
            ) from e
        time.sleep(retry_count + 1)
        return process_stream(
            stream,
            folder_path,
            mapped_variables,
            coerce_schema,
            progressbar,
            batch_preprocessor,
            cancel_event,
            offset=offset,
            retry_count=retry_count + 1,
        )



import os
import io
import warnings
from base64 import b64decode
from pathlib import Path
import time
from requests import RequestException
from urllib3.exceptions import HTTPError
from ..common import exceptions
from .Base import Base
from ..common.api_request import make_request
from urllib.parse import quote as quote_uri

from ..common.retryable_download import perform_retryable_download


class File(Base):

    def __init__(
        self,
        id,
        name,
        *,
        directory,
        table=None,
        query=None,
        properties=None,
    ):
        if not table and not query:
            raise exceptions.ValueError(
                "All files must either belong to a table or query."
            )
        if properties is None:
            properties = {}

        self.id = id
        self.path = Path(name)
        self.name = self.path.name
        self.table = table
        self.query = query
        self.directory = directory
        self.uri = f"/rawFiles/{quote_uri(id, '')}"
        self.size = properties.get("size")
        self.hash = properties.get("md5_hash")
        if self.hash:
            self.hash = b64decode(self.hash)
        self.added_at = properties.get("added_at")
        self.properties = properties

    def get(self):
        warnings.warn(
            "This method is deprecated. Nothing to fetch; all file metadata already exists at file.properties",
            FutureWarning,
            stacklevel=2,
        )
        return self

    def __repr__(self) -> str:
        return f"<File {str(self.path)}>"

    def download(
        self,
        path=None,
        *,
        overwrite=False,
        progress=True,
        on_progress=None,
        cancel_event=None,
    ):
        if path:
            path = os.path.expanduser(path)

        is_dir = False
        if path is None:
            path = os.getcwd()
            is_dir = True
        elif path.endswith(os.sep):
            is_dir = True
        elif os.path.exists(path) and os.path.isdir(path):
            is_dir = True

        query_params = {}
        if self.table:
            query_params["tableId"] = self.table.properties.get("id")
        elif self.query:
            query_params["queryId"] = self.query.properties.get("id")

        return perform_retryable_download(
            path=self.uri,
            filename=os.path.join(path, self.name) if is_dir else path,
            size=self.size,
            md5_hash=self.hash,
            query=query_params,
            progress=progress,
            on_progress=on_progress,
            cancel_event=cancel_event,
            overwrite=overwrite,
        )

    def read(self, *, as_text=False, start_byte=0, end_byte=None):
        range_headers = {}
        if end_byte:
            range_headers["Range"] = f"bytes={int(start_byte)}-{int(end_byte)}"
        elif start_byte:
            range_headers["Range"] = f"bytes={int(start_byte)}-"

        r = make_request(
            method="GET",
            path=f"{self.uri}",
            parse_response=False,
            headers=range_headers,
        )
        if as_text:
            return r.text
        else:
            return r.content

    def stream(self, *, start_byte=0, end_byte=None):
        warnings.warn(
            "file.stream() has been deprecated, please use file.open() instead",
            FutureWarning,
            stacklevel=2,
        )
        return self.open(start_byte=start_byte, end_byte=end_byte)

    def open(self, mode="r", encoding="utf-8", *, start_byte=0, end_byte=None):
        if mode not in ("r", "rb", "rt"):
            raise exceptions.ValueError(
                "Only read modes ('r', 'rb', 'rt') are supported for file.open()"
            )

        if mode == "rt":
            mode = "r"
        if start_byte:
            start_byte = int(start_byte)
        if end_byte:
            end_byte = int(end_byte)

        if mode == "r":
            return io.TextIOWrapper(
                Stream(
                    uri=self.uri, start_byte=start_byte, end_byte=end_byte, file=self
                ),
                encoding=encoding,
            )
        stream = Stream(
            uri=self.uri, start_byte=start_byte, end_byte=end_byte, file=self
        )
        return stream


class Stream(io.BufferedIOBase):
    def __init__(self, uri, start_byte=0, end_byte=None, file=None):
        super().__init__()
        self.uri = uri
        self.retry_count = 0
        self.bytes_read = 0
        self.start_byte = start_byte
        self.end_byte = end_byte
        self._file = file
        self._total_bytes = None
        self._iter_chunk_size = 1024 * 1024
        self._closed = False
        self.response = None

    def _close_response(self):
        if self.response:
            self.response.raw.close()
            self.response = None

    def _raise_if_closed(self):
        if self._closed:
            raise OSError(5, "Stream closed")

    def _get_response(self, recreate=False):
        self._raise_if_closed()

        if not recreate and self.response:
            return self.response
        try:
            # If recreating, make sure the previous response is closed
            self._close_response()

            range_headers = {}
            start_byte = self.start_byte + self.bytes_read

            if self._total_bytes is not None and start_byte >= self._total_bytes:
                # Don't attempt to read past the last byte; this will raise an error from the Redivis API
                return None

            if self.end_byte:
                range_headers["Range"] = f"bytes={start_byte}-{self.end_byte}"
            elif start_byte:
                range_headers["Range"] = f"bytes={start_byte}-"

            r = make_request(
                method="GET",
                path=self.uri,
                parse_response=False,
                stream=True,
                headers=range_headers,
            )

            self._total_bytes = self._file.size
            self.retry_count = 0
            self.response = r
            return r
        except (RequestException, HTTPError) as e:
            if self.retry_count >= 10:
                print("File download failed after too many retries, giving up.")
                raise e
            time.sleep(self.retry_count)
            self.retry_count += 1
            return self._get_response(True)

    def read(self, size=-1):
        r = self._get_response()
        if not r:
            return b""
        try:
            chunk = self.response.raw.read(size)
            self.bytes_read += len(chunk)
            return chunk
        except (RequestException, HTTPError) as e:
            if self.retry_count >= 10:
                print("File download failed after too many retries, giving up.")
                raise e
            time.sleep(self.retry_count)
            self.retry_count += 1
            self._get_response(True)
            return self.read(size)

    def read1(self, size=-1):
        return self.read(size)

    def readall(self):
        return self.read()

    def readinto(self, buffer):
        r = self._get_response()
        if not r:
            return 0
        try:
            bytes_read = self.response.raw.readinto(buffer)
            self.bytes_read += bytes_read
            return bytes_read
        except (RequestException, HTTPError) as e:
            if self.retry_count >= 10:
                print("File download failed after too many retries, giving up.")
                raise e
            time.sleep(self.retry_count)
            self.retry_count += 1
            self._get_response(True)
            return self.readinto(buffer)

    def raw(self):
        return self.response.raw

    @property
    def closed(self):
        return self._closed

    def close(self):
        self._close_response()
        self._closed = True

    def readline(self, size=-1):
        self._get_response()
        try:
            line = self.response.raw.readline(size)
            self.bytes_read += len(line)
            return line
        except (RequestException, HTTPError) as e:
            if self.retry_count >= 10:
                print("File download failed after too many retries, giving up.")
                raise e
            time.sleep(self.retry_count)
            self.retry_count += 1
            self.response = self._get_response(True)
            return self.readline(size)

    def readlines(self, hint=-1):
        self._get_response()
        if hint is None:
            hint = -1
        lines = []
        while (hint == -1 or self.bytes_read < hint) and not self._closed:
            # Try/catch handled in self.readline()
            line = self.readline()
            if line:
                lines.append(self.readline())
            else:
                break
        return lines

    def seek(self, offset, whence=os.SEEK_SET):
        self._raise_if_closed()
        current_bytes_read = self.bytes_read
        if whence == os.SEEK_SET:
            self.bytes_read = offset
        elif whence == os.SEEK_CUR:
            self.bytes_read += offset
        elif whence == os.SEEK_END:
            if self._total_bytes is None:
                self._get_response()
            self.bytes_read = self._total_bytes + offset

        if current_bytes_read != self.bytes_read:
            self._close_response()

    def readable(self):
        return True

    def seekable(self):
        return True

    def tell(self):
        return self.bytes_read

    def writable(self):
        return False

    def __iter__(self):
        return self

    def __next__(self):
        chunk = self.read(self._iter_chunk_size)
        if chunk == b"":
            self._closed = True
            self.response.close()
            raise StopIteration
        return chunk



from ..common import exceptions

from .Base import Base
from ..common.api_request import make_request


class Version(Base):
    def __init__(self, tag, *, dataset, properties=None):
        if tag != "current" and tag != "next" and not tag.lower().startswith("v"):
            tag = f"v{tag}"

        self.tag = tag
        self._dataset = dataset

        self.uri = (
            properties["uri"]
            if "uri" in (properties or {})
            else (f"{dataset.uri}/versions/{self.tag}")
        )
        self.properties = properties

    def __repr__(self):
        return f"<Version {self._dataset.qualified_reference}>"

    def get(self):
        self.properties = make_request(
            method="GET",
            path=self.uri,
        )
        self.uri = self.properties["uri"]
        return self

    def exists(self):
        try:
            make_request(method="HEAD", path=self.uri)
            return True
        except exceptions.NotFoundError:
            return False

    def update(self, *, label=None, release_notes=None):
        payload = {}
        if label is not None:
            payload["label"] = label
        if release_notes is not None:
            payload["releaseNotes"] = release_notes

        self.properties = make_request(
            method="PATCH",
            path=self.uri,
            payload=payload,
        )
        self.uri = self.properties["uri"]
        return self

    def delete(self):
        self.properties = make_request(method="DELETE", path=self.uri)
        self.uri = self.properties["uri"]

    def undelete(self):
        self.properties = make_request(method="POST", path=f"{self.uri}/undelete")
        self.uri = self.properties["uri"]

    def previous_version(self):
        if "previousVersion" not in (self.properties or {}):
            self.get()

        if not self.properties["previousVersion"]:
            return None

        return Version(
            tag=self.properties["previousVersion"]["tag"], dataset=self._dataset
        )

    def next_version(self):
        if "nextVersion" not in (self.properties or {}):
            self.get()

        if not self.properties["nextVersion"]:
            return None

        return Version(tag=self.properties["nextVersion"]["tag"], dataset=self._dataset)

    def dataset(self):
        # Avoids circular import
        from .Dataset import Dataset

        return Dataset(
            self._dataset.scopedReference,
            organization=self._dataset.organization,
            user=self._dataset.user,
            version=self.tag,
        )



import warnings
import subprocess
import sys

from .Base import Base
from .File import File
import os
from pathlib import Path
from ..common import exceptions
import concurrent.futures
from tqdm.auto import tqdm
from threading import Event
from typing import Literal, Optional, List, Union


class Directory(Base):
    def __init__(
        self,
        *,
        path: Union[str, Path],
        table: Optional["Table"] = None,
        query: Optional["Query"] = None,
        parent: Optional["Directory"] = None,
    ) -> None:
        if not table and not query:
            raise exceptions.ValueError(
                "All directories must either belong to a table or query."
            )

        self.path = path
        self.name = Path(path).name
        self.table = table
        self.query = query
        self.parent = parent
        self.children = {}
        self._mount_path = None
        self._last_cached_at = None

    def __repr__(self) -> str:
        return f"<Dir {str(self.path)}>"

    def list(
        self,
        max_results: Optional[int] = None,
        *,
        mode: Literal["all", "files", "directories"] = "all",
        recursive: bool = False,
    ) -> List[Union["Directory", File]]:
        if max_results is None:
            max_results = float("inf")
        children: List[Union["Directory", File]] = []
        # TODO: support glob pattern? or regex?
        for child in self.children.values():
            if len(children) >= max_results:
                break
            if recursive and isinstance(child, Directory):
                if mode == "all" or mode == "directories":
                    children.append(child)
                children.extend(
                    child.list(
                        mode=mode,
                        recursive=True,
                        max_results=max_results - len(children),
                    )
                )
            else:
                if (
                    mode == "all"
                    or (mode == "files" and isinstance(child, File))
                    or (mode == "directories" and isinstance(child, Directory))
                ):
                    children.append(child)
        return children

    def get(self, path: Union[str, Path]) -> Union["Directory", File, None]:
        is_explicit_dir = str(path).endswith("/")
        # This is to handle the case when the file was constructed via a file id, in legacy code
        if isinstance(path, str):
            # TODO: remove this after a bit
            split = path.split(".")
            client_id = split[0]
            if (
                len(split) == 2
                and len(client_id) == 14
                and len(client_id.split("-")) == 2
                and len(client_id.split("-")[0]) == 4
            ):
                warnings.warn(
                    "Passing file ids is deprecated, please use file names instead. E.g.: table.file('filename.png')",
                    FutureWarning,
                    stacklevel=2,
                )
                files = self.list(mode="files", recursive=True)
                for f in files:
                    if f.id == path:
                        return f
                return None

        # Normalize input to a Path
        if not isinstance(path, Path):
            path = Path(path)

        if path.is_absolute():
            # Need to use os.path.relpath instead of Path.relative_to to allow for paths that traverse above the current directory
            relative = Path(os.path.relpath(path, self.path))
        else:
            relative = path

        parts = list(relative.parts)
        if not parts:
            return self

        node = self
        for idx, part in enumerate(parts):
            # Handle current directory references
            if part == "." or part == "":
                continue

            # Handle parent directory traversal
            if part == "..":
                if node.parent is None:
                    # Can't traverse above the root of this tree
                    raise exceptions.ValueError(
                        "Path traverses above the root directory"
                    )
                node = node.parent
                # If '..' is the last segment, return the directory reached
                if idx == len(parts) - 1:
                    return node
                continue

            # Walk into children
            child = node.children.get(part)
            if not child:
                return None
            if isinstance(child, Directory):
                node = child
            else:
                # Must be the last part to match a file
                if idx == len(parts) - 1:
                    if is_explicit_dir:
                        return None
                    return child
                return None

        # If we finished traversing parts without returning, we resolved to a directory
        return node

    def mount(
        self, path: Optional[Union[str, Path]] = None, *, foreground: bool = False
    ) -> None:
        if (
            os.getenv("REDIVIS_NOTEBOOK_ID") is not None
            and os.getenv("REDIVIS_NOTEBOOK_ENABLE_FUSE") != "TRUE"
        ):
            raise RuntimeError(
                "Mounting directories is not supported within the default Redivis Notebooks. You must configure a custom machine to call directory.mount()"
            )
        from ..common.mount_directory import mount_directory

        if path is None:
            if self.parent is not None:
                default_name = self.name
            elif self.table:
                default_name = self.table.properties["name"]
            else:
                default_name = "redivis_mount"

            path = Path.cwd() / default_name
        elif isinstance(path, str):
            path = Path(path)

        mount_path = path.expanduser()
        mount_directory(self, mount_path, foreground=foreground)
        self._mount_path = mount_path
        return mount_path

    def unmount(self) -> None:
        if self._mount_path is None:
            raise exceptions.ValueError(
                "This directory is not currently mounted. Call mount() first."
            )

        mount_path = str(self._mount_path)

        try:
            if sys.platform == "darwin":
                subprocess.run(["umount", mount_path], check=True, capture_output=True)
            elif sys.platform == "win32":
                import ctypes

                kernel32 = ctypes.windll.kernel32
                if not kernel32.DefineDosDeviceW(0x00000002, mount_path, None):
                    raise OSError(f"Failed to unmount {mount_path}")
            else:
                subprocess.run(
                    ["fusermount", "-u", mount_path], check=True, capture_output=True
                )
        except subprocess.CalledProcessError as e:
            raise OSError(
                f"Failed to unmount {mount_path}: {e.stderr.decode().strip()}"
            )

        self._mount_path = None

        # The FUSE background thread also removes the directory on exit,
        # but attempt removal here as well in case that hasn't run yet.
        try:
            Path(mount_path).rmdir()
        except OSError:
            pass

        print(f"Unmounted directory at {mount_path}")

    def download(
        self,
        path: Optional[Union[str, Path]] = None,
        *,
        max_results: Optional[int] = None,
        max_parallelization: Optional[int] = None,
        overwrite: bool = False,
        progress: bool = True,
    ) -> None:
        if isinstance(path, str):
            path = Path(path)
        if path is None:
            path = Path.cwd() / self.name

        if path.exists():
            if path.is_file():
                if overwrite:
                    path.unlink()
                else:
                    raise FileExistsError(
                        f"Destination path '{path}' exists and is a file; set overwrite=True to replace it."
                    )

        path.mkdir(parents=True, exist_ok=True)
        cancel_event = Event()

        with concurrent.futures.ThreadPoolExecutor(
            max_workers=min(
                100, max_parallelization if max_parallelization is not None else 50
            ),
        ) as executor:
            futures = []
            files_to_download = self.list(
                mode="files", recursive=True, max_results=max_results
            )

            file_count = 0
            total_bytes = 0
            for f in files_to_download:
                total_bytes += f.size

            pbar = tqdm(
                unit="B",
                leave=False,
                unit_scale=True,
                mininterval=0.1,
                total=total_bytes if progress else 0,
                desc=f"{file_count}/{len(files_to_download)} files",
                disable=not progress,
            )

            def on_progress(chunk_bytes: int) -> None:
                pbar.update(chunk_bytes)

            # Create and submit download tasks with per-file progress callbacks
            for file in files_to_download:
                file_relative_path = (Path("/") / file.path).relative_to(self.path)
                dest_path = os.path.join(path, str(file_relative_path))

                futures.append(
                    executor.submit(
                        file.download,
                        path=dest_path,
                        overwrite=overwrite,
                        progress=False,
                        on_progress=on_progress if progress else None,
                        cancel_event=cancel_event,
                    )
                )

            not_done = futures
            try:
                while not_done and not cancel_event.is_set():
                    # next line 'sleeps' this main thread, letting the thread pool run
                    freshly_done, not_done = concurrent.futures.wait(
                        not_done, timeout=0.5
                    )
                    for future in freshly_done:
                        # Call result() on any finished threads to raise any exceptions encountered.
                        future.result()
                        file_count += 1
                        pbar.set_description(
                            f"{file_count}/{len(files_to_download)} files"
                        )
            except KeyboardInterrupt:
                print("KeyboardInterrupt")
            finally:
                cancel_event.set()

                for future in not_done:
                    # Only cancels futures that were never started
                    future.cancel()

                # Shutdown all background threads, now that they should know to exit early.
                executor.shutdown(wait=True, cancel_futures=True)
                pbar.close()

    def _add_file(self, file: File) -> None:
        relative_path_parts = (Path("/") / file.path).relative_to(self.path).parts

        if len(relative_path_parts) == 1:
            self.children[relative_path_parts[0]] = file
            file.directory = self
        else:
            subdir = self.children.get(str(relative_path_parts[0]))
            if not subdir:
                subdir = Directory(
                    parent=self,
                    path=self.path / relative_path_parts[0],
                    table=self.table,
                    query=self.query,
                )
                self.children[(str(relative_path_parts[0]))] = subdir
            subdir._add_file(file)



import time

from ..common import exceptions

from .Base import Base
from ..common.api_request import make_request


class Variable(Base):
    def __init__(
        self,
        name,
        *,
        table=None,
        upload=None,
        query=None,
        properties=None,
    ):
        self.name = name
        self.table = table
        self.upload = upload
        self.query = query

        if "uri" in (properties or {}):
            self.uri = properties["uri"]
        elif table:
            self.uri = f"{table.uri}/variables/{self.name}"
        elif upload:
            self.uri = f"{upload.uri}/variables/{self.name}"
        else:
            self.uri = f"{query.uri}/variables/{self.name}"

        self.properties = properties

    def __repr__(self):
        if self.properties and self.properties.get("type"):
            return f"<Variable {self.name} ({self.properties.get('type')})>"
        return f"<Variable {self.name}>"

    def get(self):
        self.properties = make_request(
            method="GET",
            path=self.uri,
        )
        self.uri = self.properties["uri"]

        return self

    def get_statistics(self):
        statistics = make_request(
            method="GET",
            path=f"{self.uri}/statistics",
        )
        while statistics["status"] == "running" or statistics["status"] == "queued":
            time.sleep(2)
            statistics = make_request(
                method="GET",
                path=f"{self.uri}/statistics",
            )

        return statistics

    def update(self, *, label=None, description=None, value_labels=None):
        payload = {}
        if label is not None:
            payload["label"] = label
        if description is not None:
            payload["description"] = description
        if value_labels is not None:
            payload["valueLabels"] = value_labels

        response = make_request(
            method="PATCH",
            path=f"{self.uri}",
            payload=payload,
        )
        self.properties = response
        return self

    def exists(self):
        try:
            make_request(method="HEAD", path=self.uri)
            return True
        except exceptions.NotFoundError:
            return False



import os
import stat
import errno
import threading

from ..common import exceptions
from mfusepy import FUSE, FuseOSError, Operations


class RedivisFS(Operations):
    def __init__(self, directory):
        self.directory = directory
        self._file_handles = {}
        self._next_fh = 1
        self._fh_lock = threading.Lock()

    def _get_node(self, path):
        """Get the file or directory node for the given path"""
        if path == "/":
            return self.directory

        # Remove leading slash and get node
        clean_path = path.lstrip("/")
        try:
            node = self.directory.get(clean_path)
        except exceptions.ValueError:
            raise FuseOSError(errno.ENOENT)

        if node is None:
            raise FuseOSError(errno.ENOENT)
        else:
            return node

    def getattr(self, path, fh=None):
        """Get file attributes"""
        node = self._get_node(path)

        # Default attributes
        attrs = {
            "st_uid": os.getuid(),
            "st_gid": os.getgid(),
            "st_atime": 0,
            "st_mtime": 0,
            "st_ctime": 0,
        }

        if hasattr(node, "children"):  # Directory
            attrs["st_mode"] = stat.S_IFDIR | 0o755
            attrs["st_nlink"] = 2
            attrs["st_size"] = 0
        else:  # File
            attrs["st_mode"] = stat.S_IFREG | 0o644
            attrs["st_nlink"] = 1
            attrs["st_size"] = node.size or 0
            if hasattr(node, "added_at") and node.added_at:
                attrs["st_mtime"] = int(node.added_at.timestamp())

        return attrs

    def readdir(self, path, fh):
        """List directory contents"""
        node = self._get_node(path)

        if not hasattr(node, "children"):
            raise FuseOSError(errno.ENOTDIR)

        entries = [".", ".."]
        for child_name in node.children.keys():
            entries.append(child_name)

        for entry in entries:
            yield entry

    def open(self, path, flags):
        """Open a file"""
        node = self._get_node(path)

        if hasattr(node, "children"):
            raise FuseOSError(errno.ENOENT)

        # Only allow read access
        if (flags & os.O_WRONLY) or (flags & os.O_RDWR):
            raise FuseOSError(errno.EACCES)

        with self._fh_lock:
            fh = self._next_fh
            self._next_fh += 1
            # Add a per-handle lock so that operations on the same handle
            # remain serialized without blocking other handles.
            self._file_handles[fh] = {
                "node": node,
                "stream": None,
                "position": 0,
                "lock": threading.Lock(),
            }

        return fh

    def read(self, path, length, offset, fh):
        """Read from a file"""
        # First, look up the handle and its per-handle lock under the
        # global file-handle lock, then release it before doing I/O.
        with self._fh_lock:
            handle = self._file_handles.get(fh)
            if handle is None:
                raise FuseOSError(errno.EBADF)
            handle_lock = handle.get("lock")

        # Fallback in case an older handle dict is missing "lock"
        if handle_lock is None:
            handle_lock = threading.Lock()
            with self._fh_lock:
                current = self._file_handles.get(fh)
                if current is None:
                    raise FuseOSError(errno.EBADF)
                if "lock" not in current:
                    current["lock"] = handle_lock
                else:
                    handle_lock = current["lock"]

        with handle_lock:
            # Snapshot the current handle state under the global lock,
            # then perform I/O without holding _fh_lock.
            with self._fh_lock:
                handle = self._file_handles.get(fh)
                if handle is None:
                    raise FuseOSError(errno.EBADF)
                node = handle["node"]
                stream = handle["stream"]
                position = handle["position"]

            try:
                # Create or reuse stream
                if not stream or position != offset:
                    if stream:
                        stream.close()
                    stream = node.open(mode="rb", start_byte=offset)
                    position = offset

                data = stream.read(length)
                position += len(data)

            except Exception:
                # Map any I/O error to a generic EIO for FUSE
                raise FuseOSError(errno.EIO)

            # Update shared handle state under the global lock, ensuring the
            # handle still exists.
            with self._fh_lock:
                current = self._file_handles.get(fh)
                if current is None:
                    # Handle was released while we were reading; close the
                    # local stream and report EBADF.
                    if stream:
                        try:
                            stream.close()
                        except Exception:
                            pass
                    raise FuseOSError(errno.EBADF)
                current["stream"] = stream
                current["position"] = position

            return data

    def release(self, path, fh):
        """Close a file"""
        # Remove the handle from the table under the global lock, but perform
        # any potentially slow close operations without holding _fh_lock.
        stream_to_close = None
        with self._fh_lock:
            handle = self._file_handles.pop(fh, None)
            if handle is not None:
                stream_to_close = handle.get("stream")

        if stream_to_close:
            try:
                stream_to_close.close()
            except Exception:
                pass
        return 0

    def statfs(self, path):
        """Get filesystem statistics"""
        return {
            "f_bsize": 4096,
            "f_blocks": 1000000,
            "f_bavail": 1000000,
            "f_bfree": 1000000,
        }


def _run_fuse_and_cleanup(fs, mount_path):
    """Run the FUSE event loop and remove the mount directory when it exits."""
    try:
        FUSE(
            fs,
            str(mount_path),
            nothreads=False,
            foreground=True,
            max_threads=os.cpu_count() or 2,
        )
    except Exception as e:
        print(e)
        pass
    finally:
        try:
            mount_path.rmdir()
        except OSError:
            pass


def mount_directory(directory, path, foreground):

    mount_path = path.expanduser()

    if mount_path.exists():
        raise exceptions.ValueError(f"Mount path {mount_path} already exists")

    mount_path.mkdir(parents=True)

    # Create and start FUSE filesystem
    fs = RedivisFS(directory)
    print(f"Mounted directory at {mount_path}")
    if foreground:
        _run_fuse_and_cleanup(fs, mount_path)
    else:
        mount_thread = threading.Thread(
            target=_run_fuse_and_cleanup,
            args=(fs, mount_path),
            daemon=True,
        )
        mount_thread.start()
        return mount_thread



from .classes.Workflow import Workflow as workflow
from .classes.Dataset import Dataset as dataset
from .classes.Datasource import Datasource as datasource
from .classes.User import User as user
from .classes.Organization import Organization as organization
from .classes.Parameter import Parameter as parameter
from .classes.Query import Query as query
from .classes.Table import Table as table
from .classes.Notebook import Notebook as notebook
from .classes.Transform import Transform as transform
from .common import exceptions
from .common.api_request import make_request as make_api_request

# Note: these should be deleted at the end to clean up the namespace
import warnings
import os

from ._version import __version__


def file(*args, **kwargs):

    raise exceptions.DeprecationError(
        'Calling redivis.file() is no longer supported. Please use redivis.table("table_reference").file("filename") instead.'
    )


def _warning(message, category, filename, lineno, file=None, line=None):
    print(f"Warning: {message}")


warnings.showwarning = _warning


def authenticate(scope=None, force_reauthentication=False):
    from .common.auth import clear_cached_credentials, get_auth_token

    if force_reauthentication:
        clear_cached_credentials()

    if isinstance(scope, str):
        scope = [scope]

    get_auth_token(scope=scope)


def current_notebook():
    import os

    if os.getenv("REDIVIS_DEFAULT_NOTEBOOK") is not None:
        return notebook(os.getenv("REDIVIS_DEFAULT_NOTEBOOK"))

    return None


def current_user():
    res = make_api_request(method="GET", path="/users/me")
    return user(res["name"], properties=res)


def current_workflow():
    import os

    if os.getenv("REDIVIS_DEFAULT_WORKFLOW") is not None:
        return workflow(os.getenv("REDIVIS_DEFAULT_WORKFLOW"))

    return None


__all__ = [
    "workflow",
    "dataset",
    "datasource",
    "user",
    "organization",
    "parameter",
    "query",
    "table",
    "notebook",
    "transform",
    "exceptions",
    "make_api_request",
    "__version__",
    "authenticate",
    "current_notebook",
    "current_user",
    "current_workflow",
    "file",
]


def _install_excepthook():
    import sys
    import traceback
    import os

    _package_dir = os.path.dirname(os.path.abspath(__file__))
    _original_hook = sys.excepthook

    def _format_filtered_tb(exc_type, exc_value, exc_tb):

        entries = traceback.extract_tb(exc_tb)

        # Extract all frames, keep only those outside the redivis package
        # Also exclude ipython noise
        # Directories to filter out of tracebacks
        _filter_dirs = [
            _package_dir,
            os.path.join("IPython", ""),
            os.path.join("ipykernel", ""),
        ]

        user_entries = [
            e
            for e in entries
            if not any(d in os.path.abspath(e.filename) for d in _filter_dirs)
        ]

        lines = []
        lines.append("Traceback (most recent call last):\n")
        if user_entries:
            lines.extend(traceback.format_list(user_entries))
        else:
            # No user frames; show the last frame before filtering so there's some context
            last_entry = entries[-1] if entries else None
            if last_entry:
                lines.extend(traceback.format_list([last_entry]))
            else:
                lines.append("  (full traceback omitted from redivis internals)\n")

        lines.append(f"{exc_type.__name__}: {exc_value}\n")
        return "".join(lines)

    def _custom_excepthook(exc_type, exc_value, exc_tb):
        if isinstance(exc_value, exceptions.RedivisError):
            print(
                _format_filtered_tb(exc_type, exc_value, exc_tb),
                file=sys.stderr,
                end="",
            )
        else:
            _original_hook(exc_type, exc_value, exc_tb)

    sys.excepthook = _custom_excepthook

    try:
        from IPython import get_ipython

        ipython = get_ipython()
        if ipython is not None:

            def _ipython_custom_exc(shell, exc_type, exc_value, exc_tb, tb_offset=None):
                try:
                    print(
                        _format_filtered_tb(exc_type, exc_value, exc_tb),
                        file=sys.stderr,
                        end="",
                    )
                except Exception as e:
                    print(
                        "An error occurred while formatting the exception:",
                        e,
                        file=sys.stderr,
                    )
                    # Fallback to original IPython exception handler
                    shell.showtraceback(
                        (exc_type, exc_value, exc_tb), tb_offset=tb_offset
                    )

            ipython.set_custom_exc((exceptions.RedivisError,), _ipython_custom_exc)
    except ImportError:
        pass


if os.getenv("REDIVIS_INSTALL_EXCEPTHOOK", "1").lower() not in ("0", "false", "no"):
    _install_excepthook()

# Register fsspec filesystem if available
try:
    from .common.fsspec import register

    register()
except (ImportError, ModuleNotFoundError):
    # fsspec or its integration is not installed; registration is optional
    pass
except Exception as exc:
    # Unexpected error during registration; warn rather than fail silently
    warnings.warn(
        f"Failed to register Redivis fsspec filesystem due to an unexpected error: {exc}",
        RuntimeWarning,
    )

# clean up namespace
del os, warnings



from tqdm.auto import tqdm
import os
import pathlib
from base64 import b64decode
import time
import hashlib
import re
from ..common import exceptions
from ..common.api_request import make_request
from contextlib import closing
from requests import RequestException
from urllib3.exceptions import HTTPError

md5_regexp = re.compile(r"(?:^|,)\s*md5\s*=\s*([^,\s]+)\s*(?=,|$)", re.IGNORECASE)


def perform_retryable_download(
    #     IMPORTANT: if updating params, make sure to update retry call signature below
    path,
    *,
    filename=None,
    progress=True,
    pbar=None,
    cancel_event=None,
    query=None,
    overwrite=False,
    size=None,
    md5_hash=None,
    start_byte=0,
    on_progress=None,
    retry_count=0,
):
    if size is not None and md5_hash is not None:
        exact_file_exists = check_filename(
            filename, overwrite, retry_count, size, md5_hash, on_progress
        )
        if exact_file_exists:
            return filename

    try:
        with closing(
            make_request(
                method="GET",
                path=path,
                stream=True,
                query=query,
                parse_response=False,
                headers={"Range": f"bytes={start_byte}-"} if start_byte else None,
            )
        ) as r:
            should_check_filename = size is None or md5_hash is None
            size = r.headers.get("x-redivis-size")

            # Looks like "crc32c=U26yZA==, md5=uE0r1xmbDXTJAGiWL6xlHw==, ..."
            md5_match = md5_regexp.search(r.headers.get("x-redivis-hash", ""))
            md5_hash = md5_match.group(1).strip() if md5_match else None

            if should_check_filename:
                exact_file_exists = check_filename(
                    filename, overwrite, retry_count, size, md5_hash, on_progress
                )
                if exact_file_exists:
                    return filename

            # Make sure output directory exists
            pathlib.Path(filename).parent.mkdir(exist_ok=True, parents=True)

            with open(filename, "wb" if retry_count == 0 else "ab") as f:
                if progress and not pbar:
                    pbar = tqdm(
                        total=int(r.headers["content-length"]),
                        leave=False,
                        unit="iB",
                        unit_scale=True,
                        mininterval=0.1,
                    )

                for chunk in r.iter_content(chunk_size=1024 * 1024):
                    retry_count = 0
                    if cancel_event and cancel_event.is_set():
                        os.remove(filename)
                        return None
                    if pbar:
                        pbar.update(len(chunk))
                    f.write(chunk)
                    if on_progress:
                        on_progress(len(chunk))
    except (RequestException, HTTPError) as e:
        if retry_count < 10:
            time.sleep(retry_count)
            return perform_retryable_download(
                path=path,
                filename=filename,
                progress=progress,
                pbar=pbar,
                query=query,
                size=size,
                md5_hash=md5_hash,
                overwrite=overwrite,
                cancel_event=cancel_event,
                start_byte=(
                    os.path.getsize(filename) if os.path.exists(filename) else 0
                ),
                on_progress=on_progress,
                retry_count=retry_count + 1,
            )
        else:
            # TODO: remove partially downloaded files
            raise exceptions.NetworkError(
                message=f"A network error occurred. Download failed after {retry_count} retries.",
                original_exception=e,
            ) from e

    if progress:
        pbar.close()

    return filename


def check_filename(filename, overwrite, retry_count, size, md5_hash, on_progress):
    if retry_count == 0 and os.path.exists(filename):
        if (
            size is not None
            and md5_hash is not None
            and os.path.getsize(filename) == int(size)
        ):
            if isinstance(md5_hash, str):
                md5_hash = b64decode(md5_hash)

            file_hash = hashlib.md5()
            with open(filename, "rb") as f:
                for byte_block in iter(lambda: f.read(8192), b""):
                    file_hash.update(byte_block)

            if file_hash.digest() == md5_hash:
                if on_progress:
                    on_progress(int(size))
                return filename
            elif not overwrite:

                raise exceptions.ValueError(
                    f"File already exists at '{filename}'. Set parameter overwrite=True to overwrite existing files."
                )
        elif not overwrite:
            raise exceptions.ValueError(
                f"File already exists at '{filename}'. Set parameter overwrite=True to overwrite existing files."
            )
    return None



from ..common import exceptions
from ..classes.Base import Base

import warnings
import os
import tempfile
from ..classes.File import File
from ..classes.Directory import Directory
from pathlib import Path
from contextlib import closing
from ..common.list_rows import list_rows
from ..common.api_request import make_request, make_paginated_request
from ..common.util import get_warning
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union, Literal
from datetime import datetime, timezone
import weakref

cached_directories: weakref.WeakValueDictionary[str, Directory] = (
    weakref.WeakValueDictionary()
)


class TabularReader(Base):
    def __init__(
        self, is_table=False, is_query=False, is_upload=False, is_read_stream=False
    ):
        self._is_table = is_table
        self._is_query = is_query
        self._is_upload = is_upload
        self._is_read_stream = is_read_stream
        self.directory = None
        self.properties = {}

    # TODO: prefix? pattern? Don't cache if these are present
    def to_directory(
        self,
        *,
        file_id_variable: Optional[str] = None,
        file_name_variable: Optional[str] = None,
    ) -> Directory:
        # TODO: add file / directory methods to uploads
        if self._is_upload:
            raise exceptions.ValueError(
                "Listing files on uploads is not currently supported"
            )
        if self._is_read_stream:
            raise exceptions.ValueError(
                "Listing files on read streams is not supported"
            )

        # We need to check that the query has finished.
        # However, for tables, we don't need to fetch the table metadata before building the directory
        #   and it's important that we don't, since the cached_directories map is based on the user-provided URI
        # TODO: in the future, we should map all URIs to a canonical URI for caching
        if self._is_query:
            check_is_ready(self)

        if self.directory is None and cached_directories.get(self.uri):
            self.directory = cached_directories[self.uri]

        if self.directory and self._is_query:
            # Queries are immutable, always use the cached version
            return self.directory

        import pyarrow

        # Compute this here, to get the timestamp before the request was sent
        now_ms = int(datetime.now(timezone.utc).timestamp() * 1000)
        headers = {}
        if self.directory:
            headers["If-Modified-Since"] = datetime.fromtimestamp(
                self.directory._last_cached_at / 1000, tz=timezone.utc
            ).strftime("%a, %d %b %Y %H:%M:%S GMT")

        with closing(
            make_request(
                method="get",
                path=f"{self.uri}/rawFiles",
                headers=headers,
                query={
                    "format": "arrow",
                    "fileIdVariable": file_id_variable,
                    "fileNameVariable": file_name_variable,
                },
                stream=True,
                parse_response=False,
            )
        ) as res:
            if res.status_code == 304:
                self.directory._last_cached_at = now_ms
                return self.directory

            directory = Directory(
                path=Path("/"),
                query=self if self._is_query else None,
                table=self if self._is_table else None,
            )

            for file_spec in (
                pyarrow.ipc.RecordBatchStreamReader(res.raw).read_all().to_pylist()
            ):
                directory._add_file(
                    File(
                        file_spec["file_id"],
                        file_spec[file_name_variable or "file_name"],
                        query=self if self._is_query else None,
                        table=self if self._is_table else None,
                        properties=file_spec,
                        directory=directory,
                    )
                )

            self.directory = directory
            self.directory._last_cached_at = now_ms
            cached_directories[self.uri] = self.directory
            return self.directory

    def file(self, path: Union[str, Path]) -> File:
        if self.directory is None and cached_directories.get(self.uri):
            self.directory = cached_directories[self.uri]

        now_ms = int(datetime.now(timezone.utc).timestamp() * 1000)
        if self.directory is None or (
            now_ms - self.directory._last_cached_at > 30000  # cache for 30s
        ):
            self.to_directory()

        node = self.directory.get(path)
        if isinstance(node, Directory):
            raise exceptions.ValueError(f"{path} is a directory, not a file")
        return node

    def list_files(
        self,
        max_results: Optional[int] = None,
        *,
        file_id_variable: Optional[str] = None,
        file_name_variable: Optional[str] = None,
    ) -> List[File]:
        self.to_directory(
            file_id_variable=file_id_variable, file_name_variable=file_name_variable
        )

        return self.directory.list(
            mode="files", recursive=True, max_results=max_results
        )

    def to_read_streams(self, target_count=os.cpu_count(), *, variables=None):
        from ..classes.ReadStream import ReadStream

        if self._is_read_stream:
            raise exceptions.ValueError(
                "Cannot call to_read_streams() on a ReadStream."
            )

        if target_count is None:
            target_count = os.cpu_count() or 1
        if target_count < 1:
            raise exceptions.ValueError("target_count must be at least 1")

        payload = {"requestedStreamCount": target_count, "format": "arrow"}

        if variables is not None:
            payload["selectedVariables"] = variables

        read_session = make_request(
            method="post",
            path=f"{self.uri}/readSessions",
            parse_response=True,
            payload=payload,
        )
        read_streams = [
            ReadStream(
                id=stream["id"],
                table=self if self._is_table else None,
                query=self if self._is_query else None,
                upload=self if self._is_upload else None,
                selected_variables=variables,
                properties=stream,
            )
            for stream in read_session["streams"]
        ]
        return read_streams

    def download_files(
        self,
        path: Optional[Union[str, Path]] = None,
        overwrite: bool = False,
        max_results: Optional[int] = None,
        file_id_variable: Optional[str] = None,
        file_name_variable: Optional[str] = None,
        progress: bool = True,
        max_parallelization: Optional[int] = None,
    ) -> None:
        warnings.warn(
            "This method is deprecated. Please use to_directory().download() instead",
            FutureWarning,
            stacklevel=2,
        )
        self.to_directory(
            file_id_variable=file_id_variable, file_name_variable=file_name_variable
        )

        return self.directory.download(
            path=path,
            max_results=max_results,
            overwrite=overwrite,
            max_parallelization=max_parallelization,
            progress=progress,
        )

    def to_arrow_dataset(
        self,
        max_results: Optional[int] = None,
        *,
        variables: Optional[Iterable[str]] = None,
        progress: bool = True,
        batch_preprocessor: Optional[Any] = None,
        max_parallelization: int = os.cpu_count() or 1,
    ) -> Any:
        mapped_variables, selected_variables, coerce_schema = get_mapped_variables(
            self, variables
        )

        return list_rows(
            uri=self.uri,
            instance=self,
            max_results=max_results,
            selected_variables=selected_variables,
            mapped_variables=mapped_variables,
            output_type="arrow_dataset",
            progress=progress,
            coerce_schema=coerce_schema,
            batch_preprocessor=batch_preprocessor,
            use_export_api=should_use_export_api(self),
            max_parallelization=max_parallelization,
        )

    def to_arrow_table(
        self,
        max_results: Optional[int] = None,
        *,
        variables: Optional[Iterable[str]] = None,
        progress: bool = True,
        batch_preprocessor: Optional[Any] = None,
        max_parallelization: int = os.cpu_count() or 1,
    ) -> Any:
        mapped_variables, selected_variables, coerce_schema = get_mapped_variables(
            self, variables
        )

        return list_rows(
            uri=self.uri,
            instance=self,
            max_results=max_results,
            selected_variables=selected_variables,
            mapped_variables=mapped_variables,
            output_type="arrow_table",
            progress=progress,
            coerce_schema=coerce_schema,
            batch_preprocessor=batch_preprocessor,
            use_export_api=should_use_export_api(self),
            max_parallelization=max_parallelization,
        )

    def to_polars_lazyframe(
        self,
        max_results: Optional[int] = None,
        *,
        variables: Optional[Iterable[str]] = None,
        progress: bool = True,
        batch_preprocessor: Optional[Any] = None,
        max_parallelization: int = os.cpu_count() or 1,
    ) -> Any:
        mapped_variables, selected_variables, coerce_schema = get_mapped_variables(
            self, variables
        )

        return list_rows(
            uri=self.uri,
            instance=self,
            max_results=max_results,
            selected_variables=selected_variables,
            mapped_variables=mapped_variables,
            output_type="polars_lazyframe",
            progress=progress,
            coerce_schema=coerce_schema,
            batch_preprocessor=batch_preprocessor,
            use_export_api=should_use_export_api(self),
            max_parallelization=max_parallelization,
        )

    def to_dask_dataframe(
        self,
        max_results: Optional[int] = None,
        *,
        variables: Optional[Iterable[str]] = None,
        progress: bool = True,
        batch_preprocessor: Optional[Any] = None,
        max_parallelization: int = os.cpu_count() or 1,
    ) -> Any:
        mapped_variables, selected_variables, coerce_schema = get_mapped_variables(
            self, variables
        )

        return list_rows(
            uri=self.uri,
            instance=self,
            max_results=max_results,
            selected_variables=selected_variables,
            mapped_variables=mapped_variables,
            output_type="dask_dataframe",
            progress=progress,
            coerce_schema=coerce_schema,
            batch_preprocessor=batch_preprocessor,
            use_export_api=should_use_export_api(self),
            max_parallelization=max_parallelization,
        )

    def to_pandas_dataframe(
        self,
        max_results: Optional[int] = None,
        *,
        variables: Optional[Iterable[str]] = None,
        progress: bool = True,
        dtype_backend: str = "pyarrow",
        date_as_object: bool = False,
        batch_preprocessor: Optional[Any] = None,
        max_parallelization: int = os.cpu_count() or 1,
    ) -> Any:
        mapped_variables, selected_variables, coerce_schema = get_mapped_variables(
            self, variables
        )

        arrow_table = list_rows(
            uri=self.uri,
            instance=self,
            max_results=max_results,
            selected_variables=selected_variables,
            mapped_variables=mapped_variables,
            output_type="arrow_table",
            progress=progress,
            coerce_schema=coerce_schema,
            batch_preprocessor=batch_preprocessor,
            use_export_api=should_use_export_api(self),
            max_parallelization=max_parallelization,
        )
        return arrow_table_to_pandas(
            arrow_table, dtype_backend, date_as_object, max_parallelization
        )

    def to_geopandas_dataframe(
        self,
        max_results: Optional[int] = None,
        *,
        variables: Optional[Iterable[str]] = None,
        geography_variable: Union[str, None, Literal[""]] = "",
        progress: bool = True,
        dtype_backend: str = "pyarrow",
        date_as_object: bool = False,
        batch_preprocessor: Optional[Any] = None,
        max_parallelization: int = os.cpu_count() or 1,
    ) -> Any:
        import geopandas

        mapped_variables, selected_variables, coerce_schema = get_mapped_variables(
            self, variables
        )

        arrow_table = list_rows(
            uri=self.uri,
            instance=self,
            max_results=max_results,
            selected_variables=selected_variables,
            mapped_variables=mapped_variables,
            output_type="arrow_table",
            progress=progress,
            coerce_schema=coerce_schema,
            batch_preprocessor=batch_preprocessor,
            use_export_api=should_use_export_api(self),
            max_parallelization=max_parallelization,
        )

        df = arrow_table_to_pandas(
            arrow_table, dtype_backend, date_as_object, max_parallelization
        )

        if geography_variable is not None:
            geography_variable = get_geography_variable(
                mapped_variables, geography_variable
            )
            if geography_variable is None:
                raise exceptions.NotFoundError(
                    'Unable to find a variable with type=="geography" in the query results'
                )

        if geography_variable is not None:
            df[geography_variable["name"]] = geopandas.GeoSeries.from_wkt(
                df[geography_variable["name"]]
            )
            df = geopandas.GeoDataFrame(
                data=df, geometry=geography_variable["name"], crs="EPSG:4326"
            )

        return df

    def to_dataframe(
        self,
        max_results: Optional[int] = None,
        *,
        variables: Optional[Iterable[str]] = None,
        geography_variable: Union[str, None, Literal[""]] = "",
        progress: bool = True,
    ) -> Any:
        warnings.warn(get_warning("dataframe_deprecation"), FutureWarning, stacklevel=2)

        mapped_variables, selected_variables, coerce_schema = get_mapped_variables(
            self, variables
        )

        arrow_table = list_rows(
            uri=self.uri,
            instance=self,
            max_results=max_results,
            selected_variables=selected_variables,
            mapped_variables=mapped_variables,
            output_type="arrow_table",
            progress=progress,
            coerce_schema=coerce_schema,
            use_export_api=should_use_export_api(self),
        )

        df = arrow_table.to_pandas(self_destruct=True)

        if geography_variable is not None:
            geography_variable = get_geography_variable(
                mapped_variables, geography_variable
            )

        if geography_variable is not None:
            import geopandas

            warnings.warn(
                get_warning("geodataframe_deprecation"), FutureWarning, stacklevel=2
            )
            df[geography_variable["name"]] = geopandas.GeoSeries.from_wkt(
                df[geography_variable["name"]]
            )
            df = geopandas.GeoDataFrame(
                data=df, geometry=geography_variable["name"], crs="EPSG:4326"
            )

        return df

    def to_arrow_batch_iterator(
        self,
        max_results: Optional[int] = None,
        *,
        variables: Optional[Iterable[str]] = None,
        progress: bool = True,
    ) -> Iterable[Any]:
        mapped_variables, selected_variables, coerce_schema = get_mapped_variables(
            self, variables
        )

        return list_rows(
            uri=self.uri,
            instance=self,
            max_results=max_results,
            selected_variables=selected_variables,
            mapped_variables=mapped_variables,
            output_type="arrow_iterator",
            progress=progress,
            coerce_schema=coerce_schema,
        )

    def to_sas(
        self,
        name: Optional[str] = None,
        max_results: Optional[int] = None,
        *,
        variables: Optional[Iterable[str]] = None,
        progress: bool = True,
        batch_preprocessor: Optional[Any] = None,
        max_parallelization: int = os.cpu_count(),
        geography_variable: Union[str, None, Literal[""]] = "",
    ) -> None:
        if self._is_read_stream:
            raise exceptions.ValueError("Cannot call to_sas() on a ReadStream.")
        check_is_ready(self)
        if not name:
            raise exceptions.ValueError(
                'A SAS dataset name must be provided. E.g., table.to_sas("mydata")'
            )
        import pyarrow as pa
        import saspy
        from IPython import get_ipython

        ip = get_ipython()

        with tempfile.TemporaryDirectory() as tmpdirname:
            # IMPORTANT: SAS is running as a separate user, need to make sure the directory is readable
            os.chmod(tmpdirname, 0o755)

            if geography_variable == "":
                # TODO: if variables param is provided, should only search within those vars
                geo_variables = make_paginated_request(
                    path=f"{self.uri}/variables",
                    query={"type": "geography"},
                    max_results=2,
                )
                if len(geo_variables) > 1:
                    raise exceptions.ValueError(
                        "Multiple geography variables found; please specify which to use"
                    )
                elif len(geo_variables) == 1:
                    geography_variable = geo_variables[0]["name"]
                else:
                    geography_variable = None

            if geography_variable is None:
                load_script = make_request(
                    method="GET",
                    path=f"{self.uri}/script",
                    query={
                        "type": "sas",
                        "filePath": f"{tmpdirname}/part-0.csv",
                        "sasDatasetName": name,
                    },
                    parse_response=False,
                ).text

                if max_results is not None and should_use_export_api(self):
                    self.download(
                        path=f"{tmpdirname}/part-0.csv", format="csv", progress=progress
                    )
                else:
                    ds = self.to_arrow_dataset(
                        max_results=max_results,
                        variables=variables,
                        progress=progress,
                        batch_preprocessor=batch_preprocessor,
                        max_parallelization=max_parallelization,
                    )
                    pa.dataset.write_dataset(
                        ds,
                        base_dir=tmpdirname,
                        existing_data_behavior="overwrite_or_ignore",
                        basename_template="part-{i}.csv",
                        format="csv",
                    )
            else:
                geopandas_df = self.to_geopandas_dataframe(
                    max_results=max_results,
                    geography_variable=geography_variable,
                    progress=progress,
                    variables=variables,
                    batch_preprocessor=batch_preprocessor,
                    max_parallelization=max_parallelization,
                )
                geopandas_df.to_file(f"{tmpdirname}/out.shp")
                load_script = f"""proc mapimport datafile="{tmpdirname}/out.shp" out={name};\nrun;"""

            ip.run_cell_magic("SAS", "", load_script)

    def to_stata(
        self,
        max_results: Optional[int] = None,
        *,
        geography_variable: Union[str, None, Literal[""]] = "",
        variables: Optional[Iterable[str]] = None,
        progress: bool = True,
        batch_preprocessor: Optional[Any] = None,
        max_parallelization: int = os.cpu_count(),
    ) -> None:
        if self._is_read_stream:
            raise exceptions.ValueError("Cannot call to_stata() on a ReadStream.")
        # This will be set if there was an error during Stata initialization in a Redivis notebook
        if os.getenv("STATA_ERROR"):
            raise exceptions.RedivisError(
                f"""An error occurred during Stata initialization. Please make sure you have the correct license and edition specified.\n\nThe error message was:\n\n{os.getenv('STATA_ERROR')}."""
            )
        check_is_ready(self)
        import pyarrow as pa
        from pystata import stata

        with tempfile.TemporaryDirectory() as tmpdirname:
            if geography_variable == "":
                # TODO: if variables param is provided, should only search within those vars
                geo_variables = make_paginated_request(
                    path=f"{self.uri}/variables",
                    query={"type": "geography"},
                    max_results=2,
                )
                if len(geo_variables) > 1:
                    raise exceptions.ValueError(
                        "Multiple geography variables found; please specify which to use"
                    )
                elif len(geo_variables) == 1:
                    geography_variable = geo_variables[0]["name"]
                else:
                    geography_variable = None

            if geography_variable is None:
                load_script_res = make_request(
                    method="GET",
                    path=f"{self.uri}/script",
                    query={"type": "stata", "filePath": f"{tmpdirname}/part-0.csv"},
                    parse_response=False,
                )
                load_script = load_script_res.text
                if max_results is not None and should_use_export_api(self):
                    self.download(
                        path=f"{tmpdirname}/part-0.csv", format="csv", progress=progress
                    )
                else:
                    ds = self.to_arrow_dataset(
                        max_results=max_results,
                        variables=variables,
                        progress=progress,
                        batch_preprocessor=batch_preprocessor,
                        max_parallelization=max_parallelization,
                    )
                    pa.dataset.write_dataset(
                        ds,
                        base_dir=tmpdirname,
                        existing_data_behavior="overwrite_or_ignore",
                        basename_template="part-{i}.csv",
                        format="csv",
                    )
            else:
                geopandas_df = self.to_geopandas_dataframe(
                    max_results=max_results,
                    geography_variable=geography_variable,
                    variables=variables,
                    progress=progress,
                    batch_preprocessor=batch_preprocessor,
                    max_parallelization=max_parallelization,
                )
                geopandas_df.to_file(f"{tmpdirname}/out.shp")
                # spshape2dta creates an out.dta + out_shp.dta file in WORKDIR, but doesn't load it. Only out.dta should be loaded; the other file is linked behind the scenes
                load_script = (
                    f'spshape2dta "{tmpdirname}/out.shp", replace\nuse out.dta, clear'
                )

            stata.run("clear")
            stata.run(load_script, quietly=True)
            stata.run("describe")

    def list_rows(
        self,
        max_results: Optional[int] = None,
        *,
        variables: Optional[Iterable[str]] = None,
        progress: bool = True,
    ):
        raise exceptions.DeprecationError(
            "The list_rows method is deprecated. Please use table.to_arrow_batch_iterator() or table.to_arrow_table().to_pylist() for better performance and memory utilization."
        )


def check_is_ready(self: TabularReader) -> None:
    if self._is_query:
        self._wait_for_finish()
    elif self._is_table:
        if not self.properties or "container" not in self.properties:
            self.get()


def get_mapped_variables(
    self: TabularReader, variables: Optional[Iterable[str]]
) -> Tuple[List[Dict[str, Any]], bool]:
    if self._is_read_stream:
        return get_mapped_variables(
            (
                self.table
                if self.table is not None
                else self.upload if self.upload is not None else self.query
            ),
            variables=self.selected_variables,
        )

    check_is_ready(self)

    coerce_schema = False  # queries and uploads will always have the correct types

    if self._is_table:
        coerce_schema = self.properties["container"]["kind"] == "dataset"

    all_variables = make_paginated_request(path=f"{self.uri}/variables", page_size=1000)

    if variables is None:
        return all_variables, variables, coerce_schema
    else:
        lower_variable_names = [variable.lower() for variable in variables]
        variables_list = list(
            filter(
                lambda variable: variable["name"].lower() in lower_variable_names,
                all_variables,
            )
        )
        variables_list.sort(
            key=lambda variable: lower_variable_names.index(variable["name"].lower())
        )
        return variables_list, variables, coerce_schema


def arrow_table_to_pandas(
    arrow_table: Any, dtype_backend: str, date_as_object: bool, max_parallelization: int
) -> Any:
    import pandas as pd
    import pyarrow as pa

    pa.set_cpu_count(max_parallelization)
    pa.set_io_thread_count(max_parallelization)

    if dtype_backend not in ["numpy", "numpy_nullable", "pyarrow"]:
        raise exceptions.ValueError(
            f"Unknown dtype_backend. Must be one of 'pyarrow'|'numpy_nullable'|'numpy'. Default is 'pyarrow'"
        )

    if dtype_backend == "numpy_nullable":
        df = arrow_table.to_pandas(
            self_destruct=True,
            date_as_object=date_as_object,
            types_mapper={
                pa.int64(): pd.Int64Dtype(),
                pa.bool_(): pd.BooleanDtype(),
                pa.float64(): pd.Float64Dtype(),
                pa.string(): pd.StringDtype(),
            }.get,
        )
    elif dtype_backend == "pyarrow":
        df = arrow_table.to_pandas(self_destruct=True, types_mapper=pd.ArrowDtype)
    else:
        df = arrow_table.to_pandas(self_destruct=True, date_as_object=date_as_object)

    return df


def get_geography_variable(
    variables: List[Dict[str, Any]],
    geography_variable_name: Union[str, None, Literal[""]],
) -> Optional[Dict[str, Any]]:
    if geography_variable_name == "":
        for variable in variables:
            if variable["type"] == "geography":
                return variable

        return None
    else:
        for variable in variables:
            if variable["name"] == geography_variable_name:
                return variable

        raise exceptions.NotFoundError(
            f"The specified geography variable '{geography_variable_name}' could not be found"
        )


def should_use_export_api(self: TabularReader) -> bool:
    if not self._is_table:
        return False
    if not self.properties or "numBytes" not in self.properties:
        self.get()
    return self.properties.get("numBytes") > (
        1e10 if os.getenv("REDIVIS_DEFAULT_NOTEBOOK") is None else 1e11
    )



import util
import redivis


def test_list_variables():
    util.populate_test_data()
    table = util.get_table()
    variables = table.list_variables()
    print(variables)


def test_list_rows():
    util.populate_test_data()
    table = util.get_table()
    print(table.list_rows(max_results=10))


def test_to_dataframe():
    util.populate_test_data()
    table = util.get_table()
    print(table.to_dataframe(max_results=10))


def test_read_streams():
    import redivis
    from concurrent.futures import ThreadPoolExecutor, as_completed

    table = redivis.table(
        "demo.cms_2014_medicare_data:349j.physicians_and_other_supplier:kn00"
    )

    streams = table.to_read_streams(
        variables=["average_submitted_chrg_amt"], target_count=4
    )

    def process_stream(stream):
        """Process a single stream, returning the sum and count for the column."""
        total = 0
        count = 0
        # Each stream can be read as an Arrow record batch reader

        for batch in stream.to_arrow_batch_iterator():
            column = batch.column("average_submitted_chrg_amt")
            # Filter out nulls
            valid = column.drop_null()
            total += valid.to_pylist().__len__() and sum(valid.to_pylist()) or 0
            count += len(valid)
        return total, count

    # Process all streams in parallel using a thread pool
    grand_total = 0
    grand_count = 0

    with ThreadPoolExecutor(max_workers=len(streams)) as executor:
        futures = {
            executor.submit(process_stream, stream): stream for stream in streams
        }
        for future in as_completed(futures):
            total, count = future.result()
            grand_total += total
            grand_count += count

    average = grand_total / grand_count if grand_count > 0 else 0
    print(f"Average submitted charge amount: ${average:.2f}")
    print(f"Computed across {grand_count} rows using {len(streams)} parallel streams")


def test_file_stream():
    from io import TextIOWrapper

    file = redivis.file("1934-e2htjwa76.AnL5YRjeUXtSlz6fhPVXog")
    lines_read = 0
    # stream = file.stream()

    # for chunk in file.stream():

    # print(stream.closed)
    with TextIOWrapper(file.stream()) as f:
        print(f.closed)
        while not f.closed:
            line = f.readline()
            if not line:
                break
            lines_read += 1
            print(lines_read)

    print(lines_read)


def test_to_arrow_batch_iterator():
    table = redivis.table("demo.ghcn_daily_weather_data.daily_observations")
    row_count = 0
    for batch in table.to_arrow_batch_iterator(1e6):
        row_count += batch.num_rows

    assert row_count == 1e6  # table.properties.get("numRows")


def test_selected_variables():
    util.populate_test_data()
    table = util.get_table()
    print(table.to_dataframe(max_results=10, variables=["ID"]))


def test_geo_dataframe():
    util.delete_test_dataset()
    util.populate_test_data("us_states.geojsonl")
    table = util.get_table()
    df = table.to_dataframe(max_results=100)
    print(df.dtypes)
    print(df)



from .Dataset import Dataset
from .Workflow import Workflow
from .Member import Member
from .Base import Base
from .Secret import Secret

from urllib.parse import quote as quote_uri

from ..common import exceptions
from ..common.api_request import make_paginated_request


class Organization(Base):
    def __init__(self, name):
        self.name = name
        self.uri = f"/organizations/{quote_uri(self.name, '')}"
        self.properties = {"kind": "organization", "uri": self.uri, "userName": name}

    def __repr__(self):
        return f"<Organization {self.name}>"

    def dataset(self, name, *, version=None):
        return Dataset(name, organization=self, version=version)

    def member(self, name):
        return Member(name, organization=self)

    def secret(self, name):
        return Secret(name, organization=self)

    def workflow(self, name):
        return Workflow(name, organization=self)

    def exists(self):
        # TODO: we don't have a get endpoint right now, so this'll do
        try:
            make_paginated_request(
                path=f"{self.uri}/datasets",
                page_size=100,
                max_results=1,
            )
            return True
        except exceptions.NotFoundError:
            return False

    def list_datasets(self, max_results=None, labels=None):
        query = {}
        if labels:
            query["labels"] = ",".join(labels)

        datasets = make_paginated_request(
            path=f"{self.uri}/datasets",
            page_size=100,
            max_results=max_results,
            query=query,
        )
        return [
            Dataset(dataset["name"], organization=self, properties=dataset)
            for dataset in datasets
        ]

    def list_members(self, max_results=None, labels=None, statuses=None):
        query = {}
        if labels:
            query["labels"] = ",".join(labels)
        if statuses:
            query["statuses"] = ",".join(statuses)

        members = make_paginated_request(
            path=f"{self.uri}/members",
            page_size=500,
            max_results=max_results,
            query=query,
        )
        return [
            Member(member["user"]["name"], organization=self, properties=member)
            for member in members
        ]

    def list_workflows(self, max_results=None):
        workflows = make_paginated_request(
            path=f"{self.uri}/workflows",
            page_size=100,
            max_results=max_results,
        )
        return [
            Workflow(workflow["name"], organization=self, properties=workflow)
            for workflow in workflows
        ]



import redivis
import util


def test_run_global_query():
    util.populate_test_data()
    query = redivis.query(
        f"SELECT * EXCEPT(_upload_name) FROM `{util.get_user_name()}.{util.get_dataset_name()}:next.{util.get_table_name()}` LIMIT 100"
    )
    print(query.list_rows(max_results=10))
    print(query.to_dataframe(max_results=10))
    assert True


def test_run_scoped_query():
    query = util.get_dataset().query(
        f"SELECT *  FROM `{util.get_table_name()}` LIMIT 100"
    )

    print(query.list_rows(max_results=10))
    print(query.to_dataframe(max_results=10))
    assert True


def test_query_list_rows():
    rows = redivis.query(
        """
            SELECT 1 + 1 AS some_number, 'foo' AS some_string
            UNION ALL 
            SELECT 4, 'bar'
        """
    ).list_rows()
    print(rows)


def test_query_empty_result():
    query = redivis.query(
        """
            SELECT * FROM (
                SELECT 1 + 1 AS val
            )sub
            WHERE FALSE
        """
    )
    print(query.to_pandas_dataframe(0))


def test_query_underscore_variable_name():
    query = redivis.query(
        """
            SELECT 1 + 1 AS _val
        """
    )
    print(query.list_rows())


def test_check_type_parsing():
    query = redivis.query(
        """
        SELECT
            1 as int,
            1.1 as float,
            'asfd' as string,
            TRUE as bool,
            CAST('2020-01-01' AS DATE) date,
            CAST('2020-01-01 12:00:00' AS DATETIME) date_time,
            CAST('01:00:00' AS TIME) time

        UNION ALL 
        SELECT NULL, NULL, NULL, NULL, NULL, NULL, NULL

        """
    )
    print(query.to_dataframe())



from .Table import Table
from .Base import Base
from ..common import exceptions
from ..common.api_request import make_request

from urllib.parse import quote as quote_uri
import os


class Datasource(Base):
    def __init__(self, source, *, workflow=None, properties=None):
        from .Workflow import Workflow
        from .Dataset import Dataset

        if not workflow:
            if os.getenv("REDIVIS_DEFAULT_WORKFLOW"):
                workflow = Workflow(os.getenv("REDIVIS_DEFAULT_WORKFLOW"))
            else:
                raise exceptions.ValueError(
                    "REDIVIS_DEFAULT_WORKFLOW must be set if no workflow is specified in the Datasource constructor"
                )
        if isinstance(source, Dataset) or isinstance(source, Workflow):
            source = source.qualified_reference

        self.workflow = workflow
        self.source_reference = source
        self.uri = (properties or {}).get(
            "uri", f"{workflow.uri}/dataSources/{quote_uri(source)}"
        )
        self.properties = properties

    def __repr__(self):
        return f"<Datasource {self.source_reference}>"

    def get(self):
        self.properties = make_request(
            method="GET",
            path=self.uri,
        )
        self.uri = self.properties["uri"]

        return self

    def create(self):
        from .Dataset import Dataset

        payload = {}
        if Dataset(self.source_reference).exists():
            payload["sourceDataset"] = self.source_reference
        else:
            payload["sourceWorkflow"] = self.source_reference

        self.properties = make_request(
            method="POST", path=f"{self.workflow.uri}/dataSources", payload=payload
        )
        self.uri = self.properties["uri"]
        return self

    def exists(self):
        try:
            make_request(method="HEAD", path=self.uri)
            return True
        except exceptions.NotFoundError:
            return False

    def source_dataset(self):
        from .Dataset import Dataset

        try:
            source_dataset = (self.properties or {})["sourceDataset"]
        except KeyError:
            self.get()
            if not "sourceDataset" in self.properties:
                raise exceptions.ValueError(
                    "This datasource doesn't have a source dataset. Use the source_workflow() method instead."
                )
            source_dataset = self.properties["sourceDataset"]
        return Dataset(source_dataset["qualifiedReference"], properties=source_dataset)

    def source_workflow(self):
        from .Workflow import Workflow

        try:
            source_workflow = (self.properties or {})["sourceWorkflow"]
        except KeyError:
            self.get()
            if not "sourceWorkflow" in self.properties:
                raise exceptions.ValueError(
                    "This datasource doesn't have a source workflow. Use the source_dataset() method instead."
                )
            source_workflow = self.properties["sourceWorkflow"]
        return Workflow(
            source_workflow["qualifiedReference"], properties=source_workflow
        )

    def update(
        self,
        *,
        source_dataset=None,
        source_workflow=None,
        sample=None,
        version=None,
        mapped_tables=None,
    ):
        # avoid circular import
        from .Workflow import Workflow
        from .Dataset import Dataset
        from .Version import Version

        if isinstance(source_workflow, Workflow):
            source_workflow = source_workflow.qualified_reference
        if isinstance(source_dataset, Dataset):
            source_dataset = source_dataset.qualified_reference
        if isinstance(mapped_tables, list):
            mapped_tables_dict = {}
            for source, dest in mapped_tables:
                if isinstance(source, Table):
                    source = source.scoped_reference
                if isinstance(dest, Table):
                    dest = dest.scoped_reference
                mapped_tables_dict[source] = dest
            mapped_tables = mapped_tables_dict

        if sample is not None or version is not None:
            if not source_dataset and not source_workflow:
                self.get()
                source_dataset = (self.properties.get("sourceDataset") or {}).get(
                    "qualifiedReference"
                )
            if not source_dataset or source_workflow:
                if sample:
                    raise exceptions.ValueError(
                        "Sampling is not applicable to datasources that reference a workflow"
                    )
                if version:
                    raise exceptions.ValueError(
                        "Versions are not applicable to datasources that reference a workflow"
                    )

        if isinstance(version, Version):
            version = version.tag

        source_dataset = Dataset(source_dataset, version=version).qualified_reference

        if sample == False:
            source_dataset.replace(":sample", "")
        elif sample == True and not ":sample" in source_dataset:
            source_dataset += ":sample"

        self.properties = make_request(
            method="PATCH",
            path=self.uri,
            payload={
                "sourceDataset": source_dataset,
                "sourceWorkflow": source_workflow,
                "mappedTables": mapped_tables,
            },
        )
        self.uri = self.properties["uri"]

        return self



import builtins


class RedivisError(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.message = message

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return self.message


class APIError(RedivisError):
    def __init__(self, message, status_code, error_description=""):
        super().__init__(error_description or message)
        self.status_code = status_code
        self.message = message
        self.description = error_description

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"[{self.status_code} {self.message}] {self.description}"


class NotFoundError(APIError):
    def __init__(self, message="", status_code=404, error_description=""):
        super().__init__(message, status_code, error_description)
        self.status_code = status_code
        self.message = message
        self.description = error_description

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return self.description or self.message


class AuthorizationError(APIError):
    def __init__(self, message, status_code, error_description=""):
        super().__init__(message, status_code, error_description)
        self.status_code = status_code
        self.message = message
        self.description = error_description

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"[{self.status_code} {self.message}] {self.description}"


class NetworkError(RedivisError):
    def __init__(
        self,
        message="A network error occurred",
        # either requests.RequestException or urllib3.exceptions.HTTPError
        original_exception=None,
    ):
        super().__init__(message)
        self.message = message
        self.original_exception = original_exception

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        if self.original_exception:
            return f"{self.message}: {self.original_exception}"
        return self.message


class ValueError(RedivisError, builtins.ValueError):
    def __init__(self, message):
        super().__init__(message)
        self.message = message

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return self.message


class JobError(RedivisError):
    def __init__(self, message=None, kind=None, status="status unknown"):
        final_message = message or f"Job finished with status: {status}"
        super().__init__(final_message)
        self.message = final_message
        self.kind = kind
        self.status = status

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"[{self.kind} {self.status}] {self.message}"


class DeprecationError(RedivisError):
    def __init__(self, message):
        super().__init__(message)
        self.message = message

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return self.message




import util

def test_patch_variable():
    util.populate_test_data()
    table = util.get_table()
    variable = table.variable('Key')
    variable.update(label="some label", description="some description", value_labels=[{ "value": "key1", "label": "a value label"}])


import os
import time
from .Variable import Variable
from ..common import exceptions

from ..common.TabularReader import TabularReader
from ..common.api_request import make_request, make_paginated_request


class Query(TabularReader):
    def __init__(
        self,
        query,
        *,
        default_workflow=None,
        default_dataset=None,
    ):
        super().__init__(is_query=True)
        self.did_initiate = False
        if not default_workflow and not default_dataset:
            if os.getenv("REDIVIS_DEFAULT_WORKFLOW"):
                default_workflow = os.getenv("REDIVIS_DEFAULT_WORKFLOW")
            elif os.getenv("REDIVIS_DEFAULT_DATASET"):
                default_dataset = os.getenv("REDIVIS_DEFAULT_DATASET")

        self.payload = {"query": query}
        self.directory = None
        if default_workflow:
            self.payload["defaultWorkflow"] = default_workflow
        if default_dataset:
            self.payload["defaultDataset"] = default_dataset

    def __repr__(self):
        return f"<Query {self.payload.get('id', '[uninitiated]')}>"

    def get(self):
        self._initiate()
        self.properties = make_request(method="GET", path=self.uri)
        return self

    # def dry_run(self):
    #   TODO

    def variable(self, name):
        # TODO: dry run (?) + cache variables
        self._wait_for_finish()
        return Variable(name, query=self)

    def list_variables(self, *, max_results=None):
        # TODO: dry run (?) + cache variables
        self._wait_for_finish()
        variables = make_paginated_request(
            path=f"{self.uri}/variables", page_size=1000, max_results=max_results
        )
        return [
            Variable(variable["name"], query=self, properties=variable)
            for variable in variables
        ]

    def _initiate(self):
        if not self.did_initiate:
            self.did_initiate = True
            self.properties = make_request(
                method="post",
                path="/queries",
                payload=self.payload,
            )
            self.uri = self.properties["uri"]

    def _wait_for_finish(self):
        self._initiate()
        while True:
            if self.properties["status"] == "completed":
                break
            elif self.properties["status"] == "failed":
                raise exceptions.JobError(
                    message=self.properties.get("errorMessage"),
                    status=self.properties.get("status"),
                    kind=self.properties.get("kind"),
                )
            elif self.properties["status"] == "cancelled":
                raise exceptions.JobError(
                    message="Query Job was cancelled",
                    status=self.properties.get("status"),
                    kind=self.properties.get("kind"),
                )
            else:
                time.sleep(2)
                self.get()



from ..common.TabularReader import TabularReader


class ReadStream(TabularReader):
    def __init__(
        self, id, *, table, query, upload, selected_variables, properties=None
    ):
        super().__init__(is_read_stream=True)
        self.id = id
        self.table = table
        self.query = query
        self.upload = upload
        self.uri = f"/readStreams/{id}"
        self.selected_variables = selected_variables
        self.properties = properties if properties is not None else {}

    def __repr__(self):
        return f"<ReadStream {self.id}>"



from fsspec import AbstractFileSystem
from ..classes.Table import Table
from ..classes.Directory import Directory
from ..classes.File import File
from . import exceptions


class RedivisFileSystem(AbstractFileSystem):
    """FSSpec filesystem implementation for Redivis."""

    protocol = "redivis"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def _parse_path(path):
        """Parse redivis://table_ref/path/to/file into (table_ref, file_path).

        The table_ref is everything before the first '/', and file_path is
        the remainder (may be empty).
        """
        if path.startswith("redivis://"):
            path = path[len("redivis://") :]
        path = path.strip("/")

        parts = path.split("/", 1)
        table_ref = parts[0]
        file_path = parts[1] if len(parts) > 1 else ""
        return table_ref, file_path

    def _get_table_and_directory(self, table_ref):
        table = Table(table_ref)
        directory = table.to_directory()  # returns root Directory
        return table, directory

    def ls(self, path, detail=False, **kwargs):
        table_ref, file_path = self._parse_path(path)
        table, root_dir = self._get_table_and_directory(table_ref)

        if file_path:
            node = root_dir.get(file_path)
        else:
            node = root_dir

        if node is None:
            raise FileNotFoundError(f"No such file or directory: {path}")

        if isinstance(node, File):
            info = self._file_info(table_ref, node)
            return [info] if detail else [info["name"]]

        entries = []
        for child in node.list():
            if isinstance(child, File):
                entries.append(self._file_info(table_ref, child))
            else:
                entries.append(
                    {
                        "name": f"{table_ref}/{child.path}",
                        "size": 0,
                        "type": "directory",
                    }
                )

        if detail:
            return entries
        return [e["name"] for e in entries]

    def _file_info(self, table_ref, file_obj):
        return {
            "name": f"{table_ref}/{file_obj.path}",
            "size": file_obj.size,
            "type": "file",
        }

    def info(self, path, **kwargs):
        table_ref, file_path = self._parse_path(path)

        if not file_path:
            return {"name": path, "size": 0, "type": "directory"}

        table, root_dir = self._get_table_and_directory(table_ref)
        node = root_dir.get(file_path)

        if node is None:
            raise FileNotFoundError(f"No such file or directory: {path}")

        if isinstance(node, Directory):
            return {"name": path, "size": 0, "type": "directory"}

        return self._file_info(table_ref, node)

    def _open(self, path, mode="rb", **kwargs):
        """Open a file from Redivis. Always returns a binary stream;
        fsspec handles text-mode wrapping externally."""
        if mode not in ("rb", "r", "rt"):
            raise exceptions.ValueError(
                f"Redivis filesystem is read-only; mode '{mode}' is not supported"
            )
        table_ref, file_path = self._parse_path(path)
        table = Table(table_ref)
        return table.file(file_path).open("rb")


def register():
    """Register the Redivis filesystem with fsspec."""
    try:
        from fsspec.registry import register_implementation

        register_implementation("redivis", RedivisFileSystem, clobber=True)
    except ImportError:
        pass

    _register_pystac_io()


def _register_pystac_io():
    """If pystac is installed, patch its default StacIO to route redivis:// URIs through fsspec."""
    try:
        import fsspec
        from pystac import StacIO
        from pystac.stac_io import DefaultStacIO

        class FsspecAwareStacIO(DefaultStacIO):
            def read_text_from_href(self, href: str) -> str:
                if href.startswith("redivis://"):
                    with fsspec.open(href, "rb") as f:
                        return f.read().decode("utf-8")
                return super().read_text_from_href(href)

            def write_text_to_href(self, href: str, txt: str) -> None:
                if href.startswith("redivis://"):
                    raise exceptions.ValueError(
                        "Writing to redivis:// is not supported"
                    )
                return super().write_text_to_href(href, txt)

        StacIO.set_default(FsspecAwareStacIO)

    except ImportError:
        pass  # pystac not installed — nothing to do



import requests
import logging
import os
import json
import platform
import warnings
from urllib.parse import unquote
import time

from .auth import get_auth_token, refresh_credentials
from .._version import __version__
from .util import raise_api_error


def make_request(
    *,
    method="GET",
    path="",
    query=None,
    payload=None,
    parse_payload=True,
    parse_response=True,
    stream=False,
    files=None,
    headers=None,
    retry_count=0,
):
    if headers is None:
        headers = {}

    original_parameters = locals().copy()
    args = get_request_args(
        method=method,
        path=path,
        query=query,
        payload=payload,
        parse_payload=parse_payload,
        stream=stream,
        files=files,
        headers=headers,
    )

    logging.debug(f"Making API '{method}' request to '{args['url']}'")
    r = requests.request(**args)

    return process_request_response(r, parse_response, method, original_parameters)


def make_paginated_request(
    *, path, query={}, page_size=100, max_results=None, parse_response=True
):
    logging.debug(f"Making paginated API request to '{path}'")

    page = 0
    results = []
    next_page_token = None

    while True:
        if max_results is not None and len(results) >= max_results:
            break

        response = make_request(
            method="get",
            path=path,
            parse_response=True,
            query={
                **query,
                **{
                    "pageToken": next_page_token,
                    "maxResults": (
                        page_size
                        if max_results is None or (page + 1) * page_size < max_results
                        else max_results - page * page_size
                    ),
                },
            },
        )
        page += 1
        results += response["results"]
        next_page_token = response["nextPageToken"]
        if not next_page_token:
            break

    return results


def get_request_args(
    method,
    path,
    query=None,
    payload=None,
    parse_payload=True,
    stream=False,
    files=None,
    headers={},
):
    api_endpoint = __get_api_endpoint()
    verify_ssl = (
        False
        if api_endpoint.find("https://localhost", 0) == 0
        or os.getenv("REDIVIS_ENV") == "development"
        or os.getenv("REDIVIS_ENV") == "test"
        or os.getenv("REDIVIS_ENV") == "staging"
        else True
    )
    url = f"{api_endpoint}{path}"

    method = method.upper()
    headers = {
        **{
            "Authorization": f"Bearer {get_auth_token()}",
            "User-Agent": f"redivis-python/{__version__} ({platform.platform()}; Python/{platform.python_version()})",
        },
        **headers,
    }

    if parse_payload and payload:
        payload = json.dumps(payload)
        headers["Content-Type"] = "application/json"

    return {
        "method": method.upper(),
        "url": url,
        "headers": headers,
        "params": query,
        "verify": verify_ssl,
        "data": payload,
        "stream": stream,
        "files": files,
        "timeout": 125,
    }


previously_printed_warnings = {}


def process_request_response(
    r, parse_response=True, method=None, original_parameters=None
):
    method = method.lower()
    response_json = {}
    try:
        # Retry with exponential backoff on service unavailable
        if r.status_code == 503 and original_parameters["retry_count"] < 0:
            logging.debug("API is currently unavailable, retrying...")
            time.sleep(original_parameters["retry_count"])
            original_parameters["retry_count"] += 1
            return make_request(**original_parameters)

        if r.status_code >= 400 or (
            method != "head" and parse_response and r.text != "OK"
        ):
            if method == "head":
                if "X-REDIVIS-ERROR-PAYLOAD" in r.headers:
                    response_json = json.loads(
                        unquote(r.headers["X-REDIVIS-ERROR-PAYLOAD"])
                    )
                else:
                    # This should never happen
                    response_json = {"error": "unknown_error", "status": r.status_code}
            else:
                response_json = r.json()

        if (
            (
                r.status_code == 401
                or (
                    r.status_code == 403
                    and response_json["error"] == "insufficient_scope"
                )
            )
            and os.getenv("REDIVIS_API_TOKEN") is None
            and os.getenv("REDIVIS_DEFAULT_NOTEBOOK") is None
        ):
            warnings.warn(
                f"{response_json['error']}: {response_json['error_description']}"
            )
            refresh_credentials(
                scope=(
                    response_json["scope"].split(" ")
                    if "scope" in response_json
                    else None
                ),
                amr_values=(
                    response_json["amr_values"]
                    if "amr_values" in response_json
                    else None
                ),
            )
            return make_request(**original_parameters)
    except Exception:
        if method == "head":
            error_payload = r.headers.get("X-REDIVIS-ERROR-PAYLOAD")
            response_text = (
                unquote(error_payload) if error_payload is not None else r.text
            )
            raise_api_error(response_text=response_text, response=r)

        else:
            raise_api_error(response_text=r.text, response=r)

    if "X-REDIVIS-WARNING" in r.headers:
        global previously_printed_warnings
        if r.headers["X-REDIVIS-WARNING"] not in previously_printed_warnings:
            warnings.warn(r.headers["X-REDIVIS-WARNING"])
            previously_printed_warnings[r.headers["X-REDIVIS-WARNING"]] = True

    if r.status_code >= 400:
        return raise_api_error(response_json=response_json, response=r)
    elif parse_response:
        return response_json
    else:
        return r


def __get_api_endpoint():
    return (
        "https://redivis.com/api/v1"
        if os.getenv("REDIVIS_API_ENDPOINT") is None
        else os.getenv("REDIVIS_API_ENDPOINT")
    )



from .Base import Base
import os
import time
import concurrent.futures
import pathlib

from ..common import exceptions
from ..common.api_request import make_request
from tqdm.auto import tqdm
import re
from threading import Event

from ..common.retryable_download import perform_retryable_download


class Export(Base):
    def __init__(
        self,
        id,
        *,
        table=None,
        properties=None,
    ):
        self.id = id
        self.table = table
        self.properties = properties
        self.uri = (self.properties or {}).get("uri") or f"/exports/{id}"

    def __repr__(self):
        return f"<Export {self.id}>"

    def get(self):
        self.properties = make_request(method="GET", path=self.uri)
        return self

    def download_files(
        self, *, path=None, overwrite=False, progress=True, max_parallelization=None
    ):
        self.wait_for_finish()
        file_count = self.properties["fileCount"]
        escaped_table_name = re.sub(
            r"\W+", "_", self.properties.get("table", {}).get("name", "table")
        ).lower()
        is_dir = False
        if path:
            path = os.path.expanduser(path)
        if path is None or (os.path.exists(path) and os.path.isdir(path)):
            is_dir = True
            if path is None:
                path = os.getcwd()
            if file_count > 1:
                path = os.path.join(path, escaped_table_name)
        elif path.endswith(os.sep) or (not os.path.exists(path) and "." not in path):
            is_dir = True
        elif file_count > 1:
            raise exceptions.ValueError(
                f"Path '{path}' is a file, but the export consists of multiple files. Please specify the path to a directory"
            )

        if (
            overwrite is False
            and os.path.exists(path)
            and (not is_dir or file_count > 1)
        ):
            raise exceptions.ValueError(
                f"File already exists at '{path}'. Set parameter overwrite=True to overwrite existing files."
            )

        # Make sure output directory exists
        if is_dir:
            pathlib.Path(path).mkdir(exist_ok=True, parents=True)
        else:
            pathlib.Path(path).parent.mkdir(exist_ok=True, parents=True)

        pbar = None
        if progress:
            pbar = tqdm(
                total=self.properties["size"],
                leave=False,
                unit="iB",
                unit_scale=True,
                mininterval=0.1,
            )

        cancel_event = Event()

        output_file_paths = []

        with concurrent.futures.ThreadPoolExecutor(
            max_workers=min(
                100, max_parallelization if max_parallelization is not None else 50
            ),
        ) as executor:
            futures = [
                executor.submit(
                    perform_retryable_download,
                    path=f"{self.uri}/download",
                    pbar=pbar,
                    cancel_event=cancel_event,
                    query={"filePart": file_number},
                    overwrite=overwrite,
                    filename=(
                        pathlib.Path(path)
                        / f"{escaped_table_name if file_count == 1 else str(file_number).zfill(6)}.{self.properties['format']}"
                        if is_dir
                        else path
                    ),
                )
                for file_number in range(file_count)
            ]

            not_done = futures

            try:
                while not_done and not cancel_event.is_set():
                    # next line 'sleeps' this main thread, letting the thread pool run
                    freshly_done, not_done = concurrent.futures.wait(
                        not_done, timeout=0.2
                    )
                    for future in freshly_done:
                        # Call result() on any finished threads to raise any exceptions encountered.
                        output_file_paths.append(future.result())
            finally:
                cancel_event.set()
                # Shutdown all background threads, now that they should know to exit early.
                executor.shutdown(wait=True, cancel_futures=True)

        if pbar:
            pbar.close()

        return output_file_paths

    def wait_for_finish(self, *, progress=True):
        iter_count = 0
        if progress:
            pbar = tqdm(
                total=100, leave=False, unit="%", unit_scale=True, mininterval=0.1
            )
            pbar.set_description(f"Preparing download...")

        while True:
            if self.properties["status"] == "completed":
                if progress:
                    pbar.close()
                break
            elif self.properties["status"] == "failed":
                if progress:
                    pbar.close()
                raise exceptions.JobError(
                    kind=self.properties.get("kind"),
                    message=self.properties.get("errorMessage"),
                    status=self.properties.get("status"),
                )
            elif self.properties["status"] == "cancelled":
                if progress:
                    pbar.close()
                raise exceptions.JobError(
                    kind=self.properties.get("kind"),
                    message=self.properties.get("errorMessage", "Export was cancelled"),
                    status=self.properties.get("status"),
                )
            else:
                iter_count += 1
                if progress:
                    pbar.update(self.properties["percentCompleted"] - pbar.n)
                time.sleep(min(iter_count * 0.5, 2))
                self.get()



from .Base import Base
from ..common import exceptions
from ..common.api_request import make_request
import os


class Parameter(Base):
    def __init__(self, name, *, workflow=None, properties=None):
        from .Workflow import Workflow  # avoid circular import

        if not workflow:
            if len(name.split(".")) == 3:
                workflow = Workflow(".".join(name.split(".")[0:2]))
                name = name.split(".")[-1]
            elif os.getenv("REDIVIS_DEFAULT_WORKFLOW"):
                workflow = Workflow(os.getenv("REDIVIS_DEFAULT_WORKFLOW"))
            else:
                raise exceptions.ValueError(
                    "Invalid parameter specifier, must be the fully qualified reference if no workflow is specified"
                )

        if isinstance(workflow, str):
            workflow = Workflow(workflow)

        self.workflow = workflow
        self.name = name

        self.properties = properties
        self.uri = (
            properties["uri"]
            if "uri" in (properties or {})
            else f"{workflow.uri}/parameters/{name}"
        )

    def __repr__(self):
        return f"<Parameter @{self.workflow.qualified_reference}.{self.name}>"

    def exists(self):
        try:
            make_request(method="HEAD", path=self.uri)
            return True
        except exceptions.NotFoundError:
            return False

    def get(self):
        self.properties = make_request(path=self.uri)
        self.uri = self.properties["uri"]
        return self

    def get_values(self):
        self.get()
        return self.properties["values"]

    def delete(self):
        make_request(path=self.uri, method="DELETE")
        return

    def update(self, *, name=None, type=None, values=None):
        payload = {}
        if name is not None:
            payload["name"] = name
        if type is not None:
            payload["type"] = type
        if values is not None:
            payload["values"] = values

        self.properties = make_request(path=self.uri, method="PATCH")
        self.uri = self.properties["uri"]
        return self

    def create(self, *, type=None, values):
        payload = {"name": self.name, "values": values}

        if type is not None:
            payload["type"] = type

        self.properties = make_request(
            path=f"{self.workflow.uri}/parameters", method="POST", payload=payload
        )
        self.uri = self.properties["uri"]
        return self



import time
import os
import json
import logging
from urllib.parse import quote as quote_uri
import warnings
import io
import pathlib
import uuid
from tqdm.auto import tqdm

from ..common import exceptions
from ..common.TabularReader import TabularReader
from ..common.util import convert_data_to_parquet

from .Variable import Variable
from ..common.api_request import make_request, make_paginated_request
from ..common.retryable_upload import perform_resumable_upload, perform_standard_upload

MAX_SIMPLE_UPLOAD_SIZE = 2**20  # 1MB
MIN_RESUMABLE_UPLOAD_SIZE = 2**25  # 32MB


class Upload(TabularReader):
    def __init__(
        self,
        name,
        *,
        table,
        properties=None,
    ):
        super().__init__(is_upload=True)
        self.table = table
        self.name = name
        self.uri = (
            properties["uri"]
            if "uri" in (properties or {})
            else (f"{table.uri}/uploads/{quote_uri(self.name.replace('.', '_'),'')}")
        )
        self.properties = properties

    def __repr__(self):
        return f"<Upload {self.table.uri}.{self.name}>"

    def delete(self):
        make_request(
            method="DELETE",
            path=self.uri,
        )

    def exists(self):
        try:
            make_request(method="HEAD", path=self.uri)
            return True
        except exceptions.NotFoundError:
            return False

    def get(self):
        self.properties = make_request(
            method="GET",
            path=self.uri,
        )
        self.uri = self.properties["uri"]

    def insert_rows(self, rows, *, update_schema=False):
        response = make_request(
            method="POST",
            path=f"{self.uri}/rows",
            payload={"rows": rows, "updateSchema": update_schema},
        )
        return response

    def list_variables(self, *, max_results=10000):
        variables = make_paginated_request(
            path=f"{self.uri}/variables", page_size=1000, max_results=max_results
        )
        return [
            Variable(variable["name"], properties=variable, upload=self)
            for variable in variables
        ]

    def variable(self, name):
        return Variable(name, upload=self)

    def create(
        self,
        content=None,
        # content can be:
        # - a file path (string or pathlib.Path)
        # - a file-like object (has a .read() method)
        # - raw bytes or a bytearray
        # - one of the data types returned by the to_ methods
        # - None (in which case, a transfer specification must be provided)
        # - TODO: a directory path (string or pathlib.Path)
        # - TODO: a Redivis Directory
        # - TODO: a Redivis File
        # - TODO: a Redivis Query (need BE support)
        # - TODO: a Redivis Table
        # - TODO: an array of any of the above, except for tables + queries
        *,
        data=None,
        type=None,  # TODO: add rawFiles (doesn't support Table or Query)
        transfer_specification=None,
        delimiter=None,
        schema=None,
        metadata=None,
        has_header_row=True,
        skip_bad_records=False,
        has_quoted_newlines=None,
        quote_character=None,
        escape_character=None,
        null_markers=None,
        rename_on_conflict=False,
        replace_on_conflict=False,
        allow_jagged_rows=False,
        if_not_exists=False,
        remove_on_fail=False,
        wait_for_finish=True,
        raise_on_fail=True,
        progress=True,
    ):

        if data is not None and content is None:
            content = data
            warnings.warn(
                "Deprecation warning: the `data` named argument has been renamed to `content`",
                FutureWarning,
                stacklevel=2,
            )

        temp_upload_id = None
        did_reopen_file = False
        tempfile_to_remove = None
        size = None

        if isinstance(content, str) and len(content) > 4096:  # maximum path length
            raise exceptions.ValueError(
                "`content` argument was a string, but is an invalid file path. To upload data as content, make sure that the `content` argument is provided as binary data."
            )

        if isinstance(content, str) or isinstance(content, pathlib.PurePath):
            did_reopen_file = True
            content = open(content, "rb")

        if hasattr(content, "name") and not self.name:
            self.name = pathlib.Path(content.name).name

        if content is not None and not hasattr(content, "read"):
            if isinstance(content, (bytes, bytearray)):
                size = len(content)
                content = io.BytesIO(content)
            else:
                tempfile_to_remove = convert_data_to_parquet(content)
                did_reopen_file = True
                type = "parquet"
                content = open(tempfile_to_remove, "rb")

        if content is not None and (
            (
                isinstance(content, io.IOBase)
                and (
                    not hasattr(content, "name")
                    or os.stat(content.name).st_size > MAX_SIMPLE_UPLOAD_SIZE
                )
            )
            or (hasattr(content, "__len__") and len(content) > MAX_SIMPLE_UPLOAD_SIZE)
        ):
            pbar_bytes = None
            # If file isn't in binary mode, we need to reopen, otherwise uploading of non-text files will fail
            if hasattr(content, "mode") and "b" not in content.mode:
                content = open(content.name, "rb")
                did_reopen_file = True

            if size is None:
                size = os.stat(content.name).st_size

            if progress:
                pbar_bytes = tqdm(
                    total=size, unit="B", leave=False, unit_scale=True, mininterval=0.1
                )

            res = make_request(
                method="POST",
                path=f"{self.table.uri}/tempUploads",
                payload={
                    "tempUploads": [
                        {
                            "size": size,
                            "name": self.name or str(uuid.uuid4()),
                            "resumable": size >= MIN_RESUMABLE_UPLOAD_SIZE,
                        }
                    ]
                },
            )
            temp_upload = res["results"][0]
            if temp_upload["resumable"]:
                perform_resumable_upload(
                    data=content,
                    size=size,
                    progressbar=pbar_bytes,
                    temp_upload_url=temp_upload["url"],
                )
            else:
                perform_standard_upload(
                    data=content,
                    temp_upload_url=temp_upload["url"],
                    progressbar=pbar_bytes,
                )

            if did_reopen_file:
                content.close()
            if progress:
                pbar_bytes.close()
            if tempfile_to_remove is not None:
                os.remove(tempfile_to_remove)

            temp_upload_id = temp_upload["id"]
            content = None

        if schema and type != "stream":
            warnings.warn(
                "The schema option is ignored for uploads that aren't of type `stream`"
            )

        exists = False

        if self.name != "":
            if self.properties is None or "uri" not in self.properties:
                self.uri = f"{self.table.uri}/uploads/{quote_uri(self.name.replace('.', '_'), '')}"
            exists = self.exists()

        if if_not_exists and exists:
            return self

        if replace_on_conflict is True and rename_on_conflict is True:
            raise exceptions.ValueError(
                "Invalid parameters. replace_on_conflict and rename_on_conflict cannot both be True."
            )

        if exists:
            if replace_on_conflict is True:
                self.delete()
            elif rename_on_conflict is not True:
                raise exceptions.ValueError(
                    f"An upload with the name {self.name} already exists on this version of the table. If you want to upload this file anyway, set the parameter rename_on_conflict=True or replace_on_conflict=True."
                )

        files = None
        payload = {
            "name": self.name,
            "type": type,
            "schema": schema,
            "metadata": metadata,
            "hasHeaderRow": has_header_row,
            "skipBadRecords": skip_bad_records,
            "hasQuotedNewlines": has_quoted_newlines,
            "allowJaggedRows": allow_jagged_rows,
            "quoteCharacter": quote_character,
            "escapeCharacter": escape_character,
            "nullMarkers": null_markers,
            "delimiter": delimiter,
            "tempUploadId": temp_upload_id,
            "transferSpecification": transfer_specification,
        }

        if content is not None:
            files = {"metadata": json.dumps(payload), "data": content}
            payload = None

        response = make_request(
            method="POST",
            path=f"{self.table.uri}/uploads",
            parse_payload=content is None,
            payload=payload,
            files=files,
        )
        self.properties = response
        self.uri = self.properties["uri"]

        try:
            if (
                content or temp_upload_id or transfer_specification
            ) and wait_for_finish:
                while True:
                    time.sleep(2)
                    self.get()
                    if (
                        self.properties["status"] == "completed"
                        or self.properties["status"] == "failed"
                    ):
                        if self.properties["status"] == "failed" and raise_on_fail:
                            raise exceptions.JobError(
                                message=self.properties.get("errorMessage"),
                                status=self.properties.get("status"),
                                kind=self.properties.get("kind"),
                            )
                        break
                    else:
                        logging.debug("Upload is still in progress...")
        except Exception:
            if remove_on_fail and self.properties["status"] == "failed":
                self.delete()
            raise

        return self



from .Base import Base
from ..common import exceptions
from ..common.api_request import make_request
from .Table import Table
import os
from urllib.parse import quote as quote_uri
import time
import logging
import warnings
from ..common.util import get_warning


class Transform(Base):
    def __init__(self, name, *, workflow=None, properties=None):
        from .Workflow import Workflow  # avoid circular import

        if not workflow:
            if len(name.split(".")) == 3:
                workflow = Workflow(".".join(name.split(".")[0:2]))
                name = name.split(".")[-1]
            elif os.getenv("REDIVIS_DEFAULT_WORKFLOW"):
                workflow = Workflow(os.getenv("REDIVIS_DEFAULT_WORKFLOW"))
            else:
                raise exceptions.ValueError(
                    "Invalid transform specifier, must be the fully qualified reference if no workflow is specified"
                )

        if isinstance(workflow, str):
            workflow = Workflow(workflow)

        self.workflow = workflow
        self.name = name

        self.qualified_reference = (properties or {}).get(
            "qualifiedReference", f"{self.workflow.qualified_reference}.{self.name}"
        )
        self.scoped_reference = (properties or {}).get("scopedReference", self.name)
        self.uri = (properties or {}).get(
            "uri", f"/transforms/{quote_uri(self.qualified_reference, '')}"
        )
        self.properties = properties

    def __repr__(self):
        return f"<Transform {self.qualified_reference}>"

    def referenced_tables(self):
        self.get()
        return [
            Table(source_table["qualifiedReference"], properties=source_table)
            for source_table in self.properties["referencedTables"]
        ]

    def source_tables(self):
        warnings.warn(
            get_warning("source_tables_deprecation"), FutureWarning, stacklevel=2
        )
        return self.referenced_tables()

    def source_table(self):
        self.get()
        return Table(
            self.properties["sourceTable"]["qualifiedReference"],
            properties=self.properties["sourceTable"],
        )

    def update(self, *, name=None, source_table=None):
        payload = {}

        if name is not None:
            payload["name"] = name
        if source_table is not None:
            if isinstance(source_table, Table):
                payload["sourceTable"] = source_table.qualified_reference
            else:
                payload["sourceTable"] = source_table

        response = make_request(
            method="PATCH",
            path=f"{self.uri}",
            payload=payload,
        )
        self.properties = response
        self.uri = self.properties["uri"]
        return self

    def output_table(self):
        self.get()
        return Table(
            self.properties["outputTable"]["qualifiedReference"],
            properties=self.properties["outputTable"],
        )

    def get(self):
        self.properties = make_request(
            method="GET",
            path=self.uri,
        )
        self.uri = self.properties["uri"]

        return self

    def exists(self):
        try:
            make_request(method="HEAD", path=self.uri)
            return True
        except exceptions.NotFoundError:
            return False

    def run(self, *, wait_for_finish=True):
        self.properties = make_request(
            method="POST",
            path=f"{self.uri}/run",
        )
        self.uri = self.properties["uri"]

        if wait_for_finish:
            while True:
                time.sleep(2)
                self.get()
                current_job = self.properties.get("currentJob") or self.properties.get(
                    "lastRunJob"
                )
                if current_job and current_job["status"] in [
                    "completed",
                    "failed",
                ]:
                    if current_job["status"] == "failed":
                        raise exceptions.JobError(
                            message=current_job.get("errorMessage"),
                            status=current_job.get("status"),
                            kind=current_job.get("kind"),
                        )
                    break
                else:
                    logging.debug("Transform is still in progress...")

        return self

    def cancel(self):
        self.properties = make_request(
            method="POST",
            path=f"{self.uri}/cancel",
        )
        self.uri = self.properties["uri"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine wheel --dev

import io
import os
import sys
import re
from shutil import rmtree

import setuptools

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = "redivis"
DESCRIPTION = "Redivis python client library"
URL = "https://github.com/redivis/redivis-python"
EMAIL = "support@redivis.com"
AUTHOR = "Redivis Inc."
REQUIRES_PYTHON = ">=3.7.0"

# Parse version string from _version.py file
VERSIONFILE = "src/redivis/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    VERSION = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

# Required dependencies
REQUIRED = [
    # Base libraries. These are required for working with the API
    "requests >= 2.0.0",
    "tqdm >= 4.64.0",
    # Data processing libraries, required if pulling data off Redivis.
    # Pyarrow is always required when working with data, other libraries only needed if returning data of that type.
    "pyarrow >= 13.0.0",
    "dask >= 2023.9.3",
    "pandas >= 2.0.0",
    "polars >= 0.19",
    "mfusepy >= 3.0.0",
    # geopandas and its dependencies
    "geopandas >= 0.14.0",
    "folium >= 0.15.0",  # required for geopandas
    "mapclassify >= 2.6.0",  # required for geopandas
    "shapely >= 2.0.1",  # required for geopandas
]

# What packages are optional?
EXTRAS = {
    # 'fancy feature': ['django'],
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, "__version__.py")) as f:
        exec(f.read(), about)
else:
    about["__version__"] = VERSION


class UploadCommand(setuptools.Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{0}".format(about["__version__"]))
        os.system("git push --tags")

        sys.exit()


# Where the magic happens:
setuptools.setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    # packages=packages,
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license="MIT",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
    ],
    # $ setup.py publish support.
    cmdclass={
        "upload": UploadCommand,
    },
)



from .Table import Table
from .Base import Base
from ..common import exceptions
from ..common.api_request import make_request
from tqdm.auto import tqdm
import time
import logging


import warnings
import pathlib
import os
from urllib.parse import quote as quote_uri

from ..common.retryable_upload import perform_resumable_upload, perform_standard_upload
from ..common.util import convert_data_to_parquet, get_warning


class Notebook(Base):
    def __init__(self, name, *, workflow=None, properties=None):
        from .Workflow import Workflow  # avoid circular import

        if not workflow:
            if len(name.split(".")) == 3:
                workflow = Workflow(".".join(name.split(".")[0:2]))
                name = name.split(".")[-1]
            elif os.getenv("REDIVIS_DEFAULT_WORKFLOW"):
                workflow = Workflow(os.getenv("REDIVIS_DEFAULT_WORKFLOW"))
            else:
                raise exceptions.ValueError(
                    "Invalid notebook specifier, must be the fully qualified reference if no workflow is specified"
                )

        if isinstance(workflow, str):
            workflow = Workflow(workflow)

        self.workflow = workflow
        self.name = name

        self.qualified_reference = (properties or {}).get(
            "qualifiedReference", f"{self.workflow.qualified_reference}.{self.name}"
        )
        self.scoped_reference = (properties or {}).get("scopedReference", self.name)
        self.uri = (properties or {}).get(
            "uri", f"/notebooks/{quote_uri(self.qualified_reference, '')}"
        )
        self.properties = properties

    def __repr__(self):
        return f"<Notebook {self.qualified_reference}>"

    def get(self):
        self.properties = make_request(
            method="GET",
            path=self.uri,
        )
        self.uri = self.properties["uri"]

        return self

    def exists(self):
        try:
            make_request(method="HEAD", path=self.uri)
            return True
        except exceptions.NotFoundError:
            return False

    def run(self, *, wait_for_finish=True):
        self.properties = make_request(
            method="POST",
            path=f"{self.uri}/run",
        )
        self.uri = self.properties["uri"]

        if wait_for_finish:
            while True:
                time.sleep(2)
                self.get()
                current_job = self.properties.get("currentJob") or self.properties.get(
                    "lastRunJob"
                )
                if current_job and current_job["status"] in [
                    "completed",
                    "failed",
                ]:
                    if current_job["status"] == "failed":
                        raise exceptions.JobError(
                            message=current_job.get("errorMessage"),
                            status=current_job.get("status"),
                            kind=current_job.get("kind"),
                        )
                    break
                else:
                    logging.debug("Notebook is still in progress...")

        return self

    def stop(self):
        self.properties = make_request(
            method="POST",
            path=f"{self.uri}/stop",
        )
        self.uri = self.properties["uri"]
        return self

    def referenced_tables(self):
        self.get()
        return [
            Table(source_table["qualifiedReference"], properties=source_table)
            for source_table in self.properties["referencedTables"]
        ]

    def source_tables(self):
        warnings.warn(
            get_warning("source_tables_deprecation"), FutureWarning, stacklevel=2
        )
        return self.referenced_tables()

    def source_table(self):
        self.get()
        return Table(
            self.properties["sourceTable"]["qualifiedReference"],
            properties=self.properties["sourceTable"],
        )

    def update(self, *, name=None, source_table=None):
        payload = {}

        if name is not None:
            payload["name"] = name
        if source_table is not None:
            if isinstance(source_table, Table):
                payload["sourceTable"] = source_table.qualified_reference
            else:
                payload["sourceTable"] = source_table

        response = make_request(
            method="PATCH",
            path=f"{self.uri}",
            payload=payload,
        )
        self.properties = response
        self.uri = self.properties["uri"]
        return self

    def output_table(self):
        self.get()
        return Table(
            self.properties["outputTable"]["qualifiedReference"],
            properties=self.properties["outputTable"],
        )

    def create_output_table(
        self, data=None, *, name=None, append=False, geography_variables=None
    ):
        if not (self.properties or {}).get("currentJob"):
            self.get()

        current_notebook_job_id = self.properties.get("currentJob")["id"]

        should_remove_tempfile = True
        temp_file_path = None
        try:
            if isinstance(data, str) or isinstance(data, pathlib.PurePath):
                data = str(data)
                if data.endswith(".parquet"):
                    should_remove_tempfile = False
                    temp_file_path = data
                elif data.endswith(".csv"):
                    import pandas

                    data = pandas.read_csv(data)
                else:
                    raise exceptions.ValueError(
                        "Only paths to parquet files (ending in .parquet) or CSV files (ending in .csv) are supported when a string argument is provided"
                    )

            if not isinstance(data, str):
                temp_file_path = convert_data_to_parquet(data)

            size = os.stat(temp_file_path).st_size
            pbar_bytes = tqdm(
                total=size, unit="B", leave=False, unit_scale=True, mininterval=0.1
            )

            res = make_request(
                method="POST",
                path=f"/notebookJobs/{current_notebook_job_id}/tempUploads",
                payload={"tempUploads": [{"size": size, "resumable": size > 5e7}]},
            )
            temp_upload = res["results"][0]

            with open(temp_file_path, "rb") as f:
                if temp_upload["resumable"]:
                    perform_resumable_upload(
                        data=f,
                        progressbar=pbar_bytes,
                        temp_upload_url=temp_upload["url"],
                    )
                else:
                    perform_standard_upload(
                        data=f,
                        temp_upload_url=temp_upload["url"],
                        progressbar=pbar_bytes,
                    )

            pbar_bytes.close()
            if should_remove_tempfile:
                os.remove(temp_file_path)

            res = make_request(
                method="PUT",
                path=f"/notebookJobs/{current_notebook_job_id}/outputTable",
                payload={
                    "name": name,
                    "append": append,
                    "geographyVariables": geography_variables,
                    "tempUploadId": temp_upload["id"],
                },
            )

            return Table(name=res["name"], properties=res)
        except Exception as e:
            if (
                temp_file_path
                and os.path.exists(temp_file_path)
                and should_remove_tempfile
            ):
                os.remove(temp_file_path)
            raise e



from urllib.parse import quote as quote_uri
import os
import time
import glob
import concurrent.futures
from tqdm.auto import tqdm


from .Upload import Upload
from .Export import Export
from .Variable import Variable
from ..common import exceptions
from ..common.TabularReader import TabularReader
from ..common.api_request import make_request, make_paginated_request
from ..common.retryable_upload import perform_resumable_upload, perform_standard_upload


class Table(TabularReader):
    def __init__(
        self,
        name,
        *,
        dataset=None,
        workflow=None,
        properties=None,
    ):
        super().__init__(is_table=True)
        if len(name.split(".")) != 3:
            dataset, workflow = get_table_parents(dataset, workflow)

        if dataset:
            reference_scope = f"{dataset.qualified_reference}."
        elif workflow:
            reference_scope = f"{workflow.qualified_reference}."
        elif len(name.split(".")) == 3:
            from .Dataset import Dataset
            from .Workflow import Workflow

            reference_scope = ".".join(name.split(".")[0:2])
            name = name.split(".")[-1]
            dataset = Dataset(reference_scope)
            workflow = Workflow(reference_scope)
            reference_scope += "."
        else:
            if not os.getenv("REDIVIS_DEFAULT_NOTEBOOK"):
                raise exceptions.ValueError(
                    "Invalid table specifier, must be the fully qualified reference if no dataset or workflow is specified"
                )
            reference_scope = ""

        self.name = name
        self.dataset = dataset
        self.workflow = workflow
        self.directory = None

        self.qualified_reference = (
            properties["qualifiedReference"]
            if "qualifiedReference" in (properties or {})
            else (f"{reference_scope}{self.name}")
        )
        self.scoped_reference = (
            properties["scopedReference"]
            if "scopedReference" in (properties or {})
            else self.name
        )
        self.uri = f"/tables/{quote_uri(self.qualified_reference, '')}"
        self.properties = properties

    def __repr__(self):
        return f"<Table {self.qualified_reference}>"

    def create(
        self, *, description=None, upload_merge_strategy="append", is_file_index=False
    ):
        payload = {
            "name": self.name,
            "uploadMergeStrategy": upload_merge_strategy,
            "isFileIndex": is_file_index,
        }
        if description is not None:
            payload["description"] = description

        response = make_request(
            method="POST",
            path=f"{self.dataset.uri}/tables",
            payload=payload,
        )
        update_properties(self, response)
        return self

    def delete(self):
        make_request(
            method="DELETE",
            path=self.uri,
        )
        return

    def get(self):
        properties = make_request(method="GET", path=self.uri)
        update_properties(self, properties)
        return self

    def exists(self):
        try:
            make_request(method="HEAD", path=self.uri)
            return True
        except exceptions.NotFoundError:
            return False

    def list_uploads(self, *, max_results=None):
        uploads = make_paginated_request(
            path=f"{self.uri}/uploads", max_results=max_results
        )
        return [
            Upload(upload["name"], table=self, properties=upload) for upload in uploads
        ]

    def list_variables(self, *, max_results=None):
        variables = make_paginated_request(
            path=f"{self.uri}/variables", page_size=1000, max_results=max_results
        )
        return [
            Variable(variable["name"], table=self, properties=variable)
            for variable in variables
        ]

    def add_files(
        self,
        *,
        files=None,
        directory=None,
        progress=True,
        max_parallelization=os.cpu_count() * 5,
    ):
        if (files is None) == (directory is None):
            raise exceptions.ValueError("Either files or directory must be specified")

        total_size = 0

        if directory:
            files = []
            if not directory.endswith("/"):
                directory += "/"
            for filename in glob.iglob(directory + "**/*", recursive=True):
                if os.path.isfile(filename):
                    size = os.stat(filename).st_size
                    total_size += size
                    files.append(
                        {
                            "path": filename,
                            "name": os.path.relpath(filename, directory),
                            "size": size,
                        }
                    )
        else:
            files = list(map(map_file, files))
            for file in files:
                total_size += file["size"]

        pbar_bytes = None
        pbar_count = None
        if progress:
            pbar_count = tqdm(
                total=len(files), leave=False, unit=" files", mininterval=0.1
            )
            pbar_bytes = tqdm(
                total=total_size,
                leave=False,
                unit="B",
                unit_scale=True,
                mininterval=0.1,
            )

        try:
            current_batch_timestamp = time.time()
            current_batch_files = []
            current_temp_uploads_batch = []
            target_batch_size = 1e8
            current_batch_size = 0

            for i, file in enumerate(files):
                if i % 1000 == 0:
                    res = make_request(
                        method="POST",
                        path=f"{self.uri}/tempUploads",
                        payload={
                            "tempUploads": [
                                {
                                    "size": file["size"],
                                    "name": file["name"],
                                    "resumable": file["size"] > 5e7,
                                }
                                for file in files[i : i + 1000]
                            ]
                        },
                    )
                    current_temp_uploads_batch = res["results"]

                current_batch_files.append(
                    {"file": file, "temp_upload": current_temp_uploads_batch[i % 1000]}
                )
                current_batch_size += file["size"]

                if (
                    len(current_batch_files) >= 1000
                    or i == len(files) - 1
                    or current_batch_size > target_batch_size
                ):

                    def upload(batch_file):
                        nonlocal pbar_bytes
                        file = batch_file["file"]
                        temp_upload = batch_file["temp_upload"]

                        if temp_upload["resumable"]:
                            data = (
                                open(file["path"], "rb")
                                if "path" in file
                                else file["data"]
                            )
                            perform_resumable_upload(
                                data=data,
                                progressbar=pbar_bytes,
                                temp_upload_url=temp_upload["url"],
                            )
                        else:
                            data = (
                                open(file["path"], "rb")
                                if "path" in file
                                else file["data"]
                            )
                            perform_standard_upload(
                                data=data,
                                temp_upload_url=temp_upload["url"],
                                progressbar=pbar_bytes,
                            )

                        if "path" in file:
                            data.close()

                    with concurrent.futures.ThreadPoolExecutor(
                        max_workers=min(100, max_parallelization)
                    ) as executor:
                        futures = [
                            executor.submit(upload, batch_file)
                            for batch_file in current_batch_files
                        ]

                        not_done = futures
                        try:
                            while not_done:
                                # next line 'sleeps' this main thread, letting the thread pool run
                                freshly_done, not_done = concurrent.futures.wait(
                                    not_done, timeout=0.5
                                )
                                for future in freshly_done:
                                    # Call result() on any finished threads to raise any exceptions encountered.
                                    future.result()
                        finally:
                            for future in not_done:
                                # Only cancels futures that were never started
                                future.cancel()
                            # Shutdown all background threads, now that they should know to exit early.
                            executor.shutdown(wait=True, cancel_futures=True)

                    if time.time() - current_batch_timestamp > 60:
                        target_batch_size = target_batch_size / 2
                    elif time.time() - current_batch_timestamp < 15:
                        target_batch_size = target_batch_size * 2

                    response = make_request(
                        method="POST",
                        path=f"{self.uri}/rawFiles",
                        payload={
                            "files": [
                                {
                                    "name": batch_file["file"]["name"],
                                    "tempUploadId": batch_file["temp_upload"]["id"],
                                }
                                for batch_file in current_batch_files
                            ]
                        },
                    )

                    if progress:
                        pbar_count.update(len(current_batch_files))

                    current_batch_timestamp = time.time()
                    current_batch_files = []
                    current_batch_size = 0

            if progress:
                pbar_bytes.close()
                pbar_count.close()
        except Exception as e:
            if progress:
                pbar_bytes.close()
                pbar_count.close()

            raise e

    def download(
        self,
        path=None,
        *,
        format="csv",
        overwrite=False,
        progress=True,
    ):
        res = make_request(
            method="POST",
            path=f"{self.uri}/exports",
            payload={"format": format},
        )
        export_job = Export(res["id"], table=self, properties=res)
        return export_job.download_files(
            path=path,
            overwrite=overwrite,
            progress=progress,
        )

    def update(self, *, name=None, description=None, upload_merge_strategy=None):
        payload = {}
        if name:
            payload["name"] = name
        if upload_merge_strategy:
            payload["uploadMergeStrategy"] = upload_merge_strategy
        if description is not None:
            payload["description"] = description

        response = make_request(
            method="PATCH",
            path=f"{self.uri}",
            payload=payload,
        )
        update_properties(self, response)
        return self

    def update_variables(self, variables):
        make_request(
            method="PATCH",
            path=f"{self.uri}/variables",
            payload={"variables": variables},
        )

    def upload(self, name=""):
        return Upload(name=name, table=self)

    def variable(self, name):
        return Variable(name, table=self)


def get_table_parents(dataset, workflow):
    from .User import User

    if dataset or workflow:
        return dataset, workflow
    elif os.getenv("REDIVIS_DEFAULT_NOTEBOOK") is not None:
        return None, None
    elif os.getenv("REDIVIS_DEFAULT_WORKFLOW") is not None:
        return None, User(os.getenv("REDIVIS_DEFAULT_WORKFLOW").split(".")[0]).workflow(
            os.getenv("REDIVIS_DEFAULT_WORKFLOW").split(".")[1]
        )
    elif os.getenv("REDIVIS_DEFAULT_DATASET") is not None:
        return (
            User(os.getenv("REDIVIS_DEFAULT_DATASET").split(".")[0]).dataset(
                os.getenv("REDIVIS_DEFAULT_DATASET").split(".")[1]
            )
        ), None
    return None, None


def update_properties(instance, properties):
    instance.properties = properties
    instance.qualified_reference = properties["qualifiedReference"]
    instance.scoped_reference = properties["scopedReference"]
    instance.name = properties["name"]
    instance.uri = properties["uri"]
    if instance.dataset and instance.workflow:
        if instance.properties.get("container"):
            if instance.properties.get("container")["kind"] == "dataset":
                instance.workflow = None
            else:
                instance.dataset = None
        elif instance.dataset.exists():
            instance.workflow = None
        else:
            instance.dataset = None


def map_file(file):
    if isinstance(file, str):
        file = {"path": file}
    else:
        file = {**file}

    if "name" not in file:
        if "data" in file:
            raise exceptions.ValueError(
                'All file specifications with a "data" key must specify a name'
            )

        file["name"] = os.path.basename(file["path"])

    if "data" in file:
        if isinstance(file["data"], str):
            file["data"] = bytes(file["data"], "utf-8")
        file["size"] = len(file["data"])
    else:
        file["size"] = os.stat(file["path"]).st_size

    return file
